"""
This is the implementation of graph-sc.
Extended from https://github.com/ciortanmadalina/graph-sc

Reference
----------
Ciortan, Madalina, and Matthieu Defrance. "GNN-based embedding for clustering scRNA-seq data." Bioinformatics 38.4 (2022): 1037-1044.
"""

import time
from tqdm import tqdm

from dgl import DGLGraph, DGLError
from dgl import function as fn
from dgl.nn.pytorch import GraphConv
from dgl.utils import expand_as_pair

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scanpy as sc
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.functional import binary_cross_entropy_with_logits as BCELoss

from sklearn.cluster import KMeans
from sklearn.decomposition import PCA
from sklearn.metrics import (adjusted_rand_score,
                             calinski_harabasz_score,
                             normalized_mutual_info_score,
                             silhouette_score)


class GraphSC(object):
    """GraphSC class.

    Parameters
    ----------
    args : argparse.Namespace
        a Namespace contains arguments of GCNAE. For details of parameters in parser args, please refer to link (parser help document).

    """

    def __init__(self, args):
        super(GraphSC, self).__init__()
        self.args = args
        self.model = GCNAE(args).to(get_device(args.use_cpu))

    def fit(self, n_epochs, dataloader, n_clusters, lr, cluster=["KMeans"]):
        """Train graph-sc.

        Parameters
        ----------
        n_epochs : int
            number of epochs.
        dataloader :
            dataloader for training.
        n_clusters : int
            number of clusters.
        lr : float
            learning rate.
        cluster : list optional
            clustering method.

        Returns
        -------
        None.

        """
        device = get_device(self.args.use_cpu)
        optim = torch.optim.Adam(self.model.parameters(), lr=lr)
        losses = []
        aris_kmeans = []
        Z = {}
        for epoch in tqdm(range(n_epochs)):
            self.model.train()
            z = []
            y = []
            order = []

            for input_nodes, output_nodes, blocks in dataloader:
                blocks = [b.to(device) for b in blocks]
                input_features = blocks[0].srcdata['features']
                g = blocks[-1]
                degs = g.in_degrees().float()

                adj_logits, emb = self.model.forward(blocks, input_features)
                z.extend(emb.detach().cpu().numpy())
                if "label" in blocks[-1].dstdata:
                    y.extend(blocks[-1].dstdata["label"].cpu().numpy())
                order.extend(blocks[-1].dstdata["order"].cpu().numpy())

                adj = g.adjacency_matrix().to_dense()
                adj = adj[g.dstnodes()]
                pos_weight = torch.Tensor(
                    [float(adj.shape[0] * adj.shape[0] - adj.sum()) / adj.sum()])
                factor = float((adj.shape[0] * adj.shape[0] - adj.sum()) * 2)
                if factor == 0:
                    factor = 1
                norm = adj.shape[0] * adj.shape[0] / factor
                adj_logits, _ = self.model.forward(blocks, input_features)
                loss = norm * BCELoss(adj_logits, adj.to(device),
                                      pos_weight=pos_weight.to(device))
                optim.zero_grad()
                loss.backward()
                optim.step()
                losses.append(loss.item())

            z = np.array(z)
            y = np.array(y)
            order = np.array(order)
            order = np.argsort(order)
            z = z[order]
            y = y[order]
            if pd.isnull(y[0]):
                y = None
            self.z = z

            if self.args.eval_epoch:
                score = self.score(y, n_clusters, cluster=cluster)
                aris_kmeans.append(score["kmeans_ari"])
                if self.args.show_epoch_ari:
                    print(f'epoch {epoch}, ARI {score.get("kmeans_ari")}')
                z_ = {f'epoch{epoch}': z}
                Z = {**Z, **z_}

            elif epoch == n_epochs - 1:
                self.z = z

        if self.args.eval_epoch:
            index = np.argmax(aris_kmeans)
            self.z = Z[f'epoch{index}']

    def predict(self, n_clusters, cluster=["KMeans"]):
        """Get predictions from the graph autoencoder model.

        Parameters
        ----------
        n_clusters : int
            number of clusters.
        cluster : list optional
            clustering method.

        Returns
        -------
        pred : dict
            prediction of given clustering method.

        """
        z = self.z
        device = get_device(self.args.use_cpu)
        pred = {}

        if "KMeans" in cluster:
            kmeans = KMeans(n_clusters=n_clusters,
                            init="k-means++", random_state=5)
            kmeans_pred = {"kmeans_pred": kmeans.fit_predict(z)}
            pred = {**pred, **kmeans_pred}

        if "Leiden" in cluster:
            leiden_pred = {"leiden_pred": run_leiden(z)}
            pred = {**pred, **leiden_pred}

        return pred

    def score(self, y, n_clusters, plot=False, cluster=["KMeans"]):
        """Evaluate the graph autoencoder model.

        Parameters
        ----------
        y : list
            true labels.
        n_clusters : int
            number of clusters.
        plot : bool optional
            show plot or not.
        cluster : list optional
            clustering method.

        Returns
        -------
        scores : dict
            metric evaluation scores.

        """
        z = self.z
        device = get_device(self.args.use_cpu)
        self.model.eval()

        k_start = time.time()
        scores = {"ae_end": k_start}

        if "KMeans" in cluster:
            kmeans = KMeans(n_clusters=n_clusters,
                            init="k-means++", random_state=5)
            kmeans_pred = kmeans.fit_predict(z)
            ari_k = None
            nmi_k = None
            if y is not None:
                ari_k = round(adjusted_rand_score(y, kmeans_pred), 4)
                nmi_k = round(normalized_mutual_info_score(y, kmeans_pred), 4)
            sil_k = silhouette_score(z, kmeans_pred)
            cal_k = calinski_harabasz_score(z, kmeans_pred)
            k_end = time.time()
            scores_k = {
                "kmeans_ari": ari_k,
                "kmeans_nmi": nmi_k,
                "kmeans_sil": sil_k,
                "kmeans_cal": cal_k,
                "kmeans_pred": kmeans_pred,
                "kmeans_time": k_end - k_start,
            }
            scores = {**scores, **scores_k}

        if "Leiden" in cluster:
            l_start = time.time()
            leiden_pred = run_leiden(z)
            ari_l = None
            nmi_l = None
            if y is not None:
                ari_l = round(adjusted_rand_score(y, leiden_pred), 4)
                nmi_l = round(normalized_mutual_info_score(y, leiden_pred), 4)
            sil_l = silhouette_score(z, leiden_pred)
            cal_l = calinski_harabasz_score(z, leiden_pred)
            l_end = time.time()
            scores_l = {
                "leiden_ari": ari_l,
                "leiden_nmi": nmi_l,
                "leiden_sil": sil_l,
                "leiden_cal": cal_l,
                "leiden_pred": leiden_pred,
                "leiden_time": l_end - l_start,
            }
            scores = {**scores, **scores_l}

        if plot:
            pca = PCA(2).fit_transform(z)
            plt.figure(figsize=(12, 4))
            plt.subplot(131)
            plt.title("Ground truth")
            plt.scatter(pca[:, 0], pca[:, 1], c=y, s=4)

            plt.subplot(132)
            plt.title("K-Means pred")
            plt.scatter(pca[:, 0], pca[:, 1], c=kmeans_pred, s=4)

            plt.subplot(133)
            plt.title("Leiden pred")
            plt.scatter(pca[:, 0], pca[:, 1], c=leiden_pred, s=4)
            plt.show()
        return scores


class GCNAE(nn.Module):
    """Graph convolutional autoencoder class.

    Parameters
    ----------
    args : argparse.Namespace
        a Namespace contains arguments of scDSC. For details of parameters in parser args, please refer to link (parser help document).
    agg : str
        aggregation layer.
    activation :str
        activation function.
    in_feats : int
        dimension of input feature
    n_hidden : int
        number of hidden layer.
    hidden_dim :int
        input dimension of hidden layer 1.
    hidden_1 : int
        output dimension of hidden layer 1.
    hidden_2 : int
        output dimension of hidden layer 2.
    dropout : float
        dropout rate.
    n_layers :int
        number of graph convolutional layers.
    hidden_relu : bool
        use relu activation in hidden layers or not.
    hidden_bn : bool
        use batch norm in hidden layers or not.

    Returns
    -------
    adj_rec :
        reconstructed adjacency matrix.
    x :
        embedding.
    """

    def __init__(self, args):

        super(GCNAE, self).__init__()
        self.args = args
        self.agg = args.agg

        if args.activation == 'gelu':
            activation = F.gelu
        elif args.activation == 'prelu':
            activation = F.prelu
        elif args.activation == 'relu':
            activation = F.relu
        elif args.activation == 'leaky_relu':
            activation = F.leaky_relu

        if args.n_hidden == 0:
            hidden = None
        elif args.n_hidden == 1:
            hidden = [args.hidden_1]
        elif args.n_hidden == 2:
            hidden = [args.hidden_1, args.hidden_2]

        if args.dropout != 0:
            self.dropout = nn.Dropout(p=args.dropout)
        else:
            self.dropout = None

        self.layer1 = WeightedGraphConv(
            in_feats=args.in_feats, out_feats=args.hidden_dim, activation=activation)
        if args.n_layers == 2:
            self.layer2 = WeightedGraphConv(
                in_feats=args.hidden_dim, out_feats=args.hidden_dim, activation=activation)
        self.decoder = InnerProductDecoder(activation=lambda x: x)
        self.hidden = hidden
        if hidden is not None:
            enc = []
            for i, s in enumerate(hidden):
                if i == 0:
                    enc.append(nn.Linear(args.hidden_dim, hidden[i]))
                else:
                    enc.append(nn.Linear(hidden[i - 1], hidden[i]))
                if args.hidden_bn and i != len(hidden):
                    enc.append(nn.BatchNorm1d(hidden[i]))
                if args.hidden_relu and i != len(hidden):
                    enc.append(nn.ReLU())
            self.encoder = nn.Sequential(*enc)

    def forward(self, blocks, features):
        x = blocks[0].srcdata['features']
        for i in range(len(blocks)):
            with blocks[i].local_scope():
                if self.dropout is not None:
                    x = self.dropout(x)
                blocks[i].srcdata['h'] = x
                if i == 0:
                    x = self.layer1(blocks[i], x, agg=self.agg)
                else:
                    x = self.layer2(blocks[i], x, agg=self.agg)
        if self.hidden is not None:
            x = self.encoder(x)
        adj_rec = self.decoder(x)
        return adj_rec, x


class InnerProductDecoder(nn.Module):
    """Inner product decoder class.

    Parameters
    ----------
    activation : optional
        activation function.
    dropout : float optional
        dropout rate.

    Returns
    -------
    adj :
        reconstructed adjacency matrix.
    """

    def __init__(self, activation=torch.sigmoid, dropout=0.1):
        super(InnerProductDecoder, self).__init__()
        self.dropout = dropout
        self.activation = activation

    def forward(self, z):
        z = F.dropout(z, self.dropout)
        adj = self.activation(torch.mm(z, z.t()))
        return adj


class WeightedGraphConv(GraphConv):
    """Adaptation of the dgl GraphConv model to use edge weights.
    """

    def edge_selection_simple(self, edges):
        """Edge selection.

        Parameters
        ----------
        edges :
            edges of graph.

        """
        return {'m': edges.src['h'] * edges.data['weight']}

    def forward(self, graph, feat, weight=None, agg="sum"):
        with graph.local_scope():
            if not self._allow_zero_in_degree:
                if (graph.in_degrees() == 0).any():
                    raise DGLError('There are 0-in-degree nodes in the graph, '
                                   'output for those nodes will be invalid. '
                                   'This is harmful for some applications, '
                                   'causing silent performance regression. '
                                   'Adding self-loop on the input graph by '
                                   'calling `g = dgl.add_self_loop(g)` will resolve '
                                   'the issue. Setting ``allow_zero_in_degree`` '
                                   'to be `True` when constructing this module will '
                                   'suppress the check and let the code run.')

            # (BarclayII) For RGCN on heterogeneous graphs we need to support GCN on bipartite.
            feat_src, feat_dst = expand_as_pair(feat, graph)
            if self._norm == 'both':
                degs = graph.out_degrees().float().clamp(min=1)
                norm = torch.pow(degs, -0.5)
                shp = norm.shape + (1,) * (feat_src.dim() - 1)
                norm = torch.reshape(norm, shp)
                feat_src = feat_src * norm

            if weight is not None:
                if self.weight is not None:
                    raise DGLError('External weight is provided while at the same time the'
                                   ' module has defined its own weight parameter. Please'
                                   ' create the module with flag weight=False.')
            else:
                weight = self.weight

            if weight is not None:
                feat_src = torch.matmul(feat_src, weight)
            graph.srcdata['h'] = feat_src
            if agg == "sum":
                graph.update_all(self.edge_selection_simple,
                                 fn.sum(msg='m', out='h'))
            if agg == "mean":
                graph.update_all(self.edge_selection_simple,
                                 fn.mean(msg='m', out='h'))
            rst = graph.dstdata['h']
            if self._norm != 'none':

                degs = graph.in_degrees().float().clamp(min=1)
                if self._norm == 'both':
                    norm = torch.pow(degs, -0.5)
                else:
                    norm = 1.0 / degs
                shp = norm.shape + (1,) * (feat_dst.dim() - 1)
                norm = torch.reshape(norm, shp)
                rst = rst * norm

            if self.bias is not None:
                rst = rst + self.bias

            if self._activation is not None:
                rst = self._activation(rst)

            return rst


class WeightedGraphConvAlpha(GraphConv):
    """Adaptation of the dgl GraphConv model to learn the extra edge weight parameter.
    """

    def edge_selection_simple(self, edges):
        """Edge selection.

        Parameters
        ----------
        edges :
            edges of graph.

        """
        number_of_edges = edges.src['h'].shape[0]
        indices = np.expand_dims(
            np.array([self.gene_num + 1] * number_of_edges, dtype=np.int32), axis=1)
        src_id, dst_id = edges.src['id'].cpu(
        ).numpy(), edges.dst['id'].cpu().numpy()
        indices = np.where((src_id >= 0) & (dst_id < 0),
                           src_id, indices)  # gene->cell
        indices = np.where((dst_id >= 0) & (src_id < 0),
                           dst_id, indices)  # cell->gene
        indices = np.where((dst_id >= 0) & (src_id >= 0),
                           self.gene_num, indices)  # gene-gene
        h = edges.src['h'] * self.alpha[indices.squeeze()]
        return {'m': h}

    #         return {'m': h * edges.data['weight']}

    def forward(self, graph, feat, weight=None, alpha=None, gene_num=None):
        self.alpha = alpha
        self.gene_num = gene_num
        with graph.local_scope():
            if not self._allow_zero_in_degree:
                if (graph.in_degrees() == 0).any():
                    raise DGLError('There are 0-in-degree nodes in the graph, '
                                   'output for those nodes will be invalid. '
                                   'This is harmful for some applications, '
                                   'causing silent performance regression. '
                                   'Adding self-loop on the input graph by '
                                   'calling `g = dgl.add_self_loop(g)` will resolve '
                                   'the issue. Setting ``allow_zero_in_degree`` '
                                   'to be `True` when constructing this module will '
                                   'suppress the check and let the code run.')

            # (BarclayII) For RGCN on heterogeneous graphs we need to support GCN on bipartite.
            feat_src, feat_dst = expand_as_pair(feat, graph)
            #             print(f"feat_src : {feat_src.shape}, feat_dst {feat_dst.shape}")
            if self._norm == 'both':
                degs = graph.out_degrees().float().clamp(min=1)
                norm = torch.pow(degs, -0.5)
                shp = norm.shape + (1,) * (feat_src.dim() - 1)
                norm = torch.reshape(norm, shp)
                feat_src = feat_src * norm

            if weight is not None:
                if self.weight is not None:
                    raise DGLError('External weight is provided while at the same time the'
                                   ' module has defined its own weight parameter. Please'
                                   ' create the module with flag weight=False.')
                else:
                    feat_src = torch.matmul(feat_src, weight)
            else:
                weight = self.weight

            graph.srcdata['h'] = feat_src
            graph.update_all(self.edge_selection_simple,
                             fn.sum(msg='m', out='h'))
            rst = graph.dstdata['h']

            if self._norm != 'none':

                degs = graph.in_degrees().float().clamp(min=1)
                if self._norm == 'both':
                    norm = torch.pow(degs, -0.5)
                else:
                    norm = 1.0 / degs
                shp = norm.shape + (1,) * (feat_dst.dim() - 1)
                norm = torch.reshape(norm, shp)
                rst = rst * norm

            if self.bias is not None:
                rst = rst + self.bias

            if self._activation is not None:
                rst = self._activation(rst)

            return rst


def run_leiden(data):
    """Performs Leiden community detection on given data.

    Parameters
    ----------
    data :
        data for leiden

    Returns
    -------
    pred : list
        prediction of leiden.
    """
    adata = sc.AnnData(data)
    sc.pp.neighbors(adata, use_rep='X', n_neighbors=300, n_pcs=0)
    sc.tl.leiden(adata)
    pred = adata.obs['leiden'].to_list()
    pred = [int(x) for x in pred]
    return pred


def get_device(use_cpu=False):
    """Get device for training

    Parameters
    ----------
    use_cpu : bool optional
        use cpu or not.

    Returns
    -------
    device :
        torch device.
    """
    if torch.cuda.is_available() and use_cpu == False:
        device = torch.device('cuda')
    else:
        device = torch.device('cpu')
    return device


import pandas as pd
import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns 
import scanpy as sc
import scipy as sp
import numpy as np
import argparse
# Loompy is only needed if using loom files
# import loompy 
import anndata as ad
from dance.datasets.singlemodality import CellTypeDataset
from dance.modules.single_modality.cell_type_annotation.singlecellnet import SingleCellNet
import ipdb


sc.settings.verbosity = 3 
sc.logging.print_header()


if __name__ == '__main__':
    
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--data_type", type=str, default='singlecellnet')
    parser.add_argument("--example", type=bool, default=False, help="Whether this is a example based on Anndata")
    parser.add_argument("--random_seed", type=int, default=10)
    parser.add_argument("--train_dataset", default=4682, type=int, help="train id")
    parser.add_argument("--test_dataset", default=203, type=int, help="test id")
    parser.add_argument("--species", default='mouse', type=str)
    parser.add_argument("--tissue", default='Kidney', type=str)
    parser.add_argument("--train_dir", type=str, default='train')
    parser.add_argument("--test_dir", type=str, default='test')
    parser.add_argument("--proj_path", type=str, default='./')
    parser.add_argument("--map_path", type=str, default='map/mouse/')
    parser.add_argument("--nTopGenes", type=int, default=100)
    parser.add_argument("--nRand", type=int, default=100)
    parser.add_argument("--nTrees", type=int, default=100)
    parser.add_argument("--nTopGenePairs", type=int, default=250)
    parser.add_argument("--dLevel", type=str, default='Cell_type',help = "name for the cell type information for training data") 
    parser.add_argument("--dtype", type=str, default='Cell_type',help = "name for the cell type information for test data") 
    parser.add_argument("--limitToHVG", type=bool, default=True)
    parser.add_argument("--stratify", type=bool, default=True)
    args = parser.parse_args()
    
    if args.example:
        dataloader = CellTypeDataset(data_type = "singlecellnet_exp")
        dataloader = dataloader.load_data()
        model = SingleCellNet()
        model.fit(dataloader.expTrain, nTopGenes = args.nTopGenes, nRand = args.nRand, nTrees = args.nTrees ,nTopGenePairs = args.nTopGenePairs, dLevel = "cell_ontology_class", stratify=args.stratify, limitToHVG=args.limitToHVG)
        adVal = model.predict(dataloader.expVal)
        correct_pred, accuracy = model.score_exp(adVal, dtype = "cell_ontology_class")
        print(accuracy)

    else:
        
        dataloader = CellTypeDataset(random_seed = args.random_seed, data_type = "singlecellnet", proj_path = args.proj_path, train_dir = args.train_dir,test_dir = args.test_dir,train_dataset = args.train_dataset, test_dataset = args.test_dataset, species = args.species, tissue = args.tissue, map_path =  args.map_path)
        dataloader = dataloader.load_data()
        
        model = SingleCellNet()
        model.fit(dataloader.train_adata, nTopGenes = args.nTopGenes, nRand = args.nRand, nTrees = args.nTrees ,nTopGenePairs = args.nTopGenePairs, dLevel = args.dLevel, stratify=args.stratify, limitToHVG=args.limitToHVG)
      
        adVal = model.predict(dataloader.test_adata)
        ipdb.set_trace()
        acc = model.score(adNew = adVal, test_dataset= dataloader.test_adata, map = dataloader.map_dict[args.test_dataset],dtype = args.dtype)
        print(acc) 

"""To reproduce SingleCellNet benchmarks, please refer to command lines belows:

Mouse Brain
$ python singlecellnet.py --species mouse --tissue Brain --train_dataset 753 --test_dataset 2695 --dLevel Cell_type --dtype Cell_type

Mouse Spleen
$ python singlecellnet.py --species mouse --tissue Spleen --train_dataset 1970 --test_dataset 1759 --dLevel Cell_type --dtype Cell_type

Mouse Kidney
$ python singlecellnet.py --species mouse --tissue Kidney --train_dataset 4682 --test_dataset 203 --dLevel Cell_type --dtype Cell_type

"""



"""
Main functionality for starting training.
This code is based on https://github.com/NVlabs/MUNIT.
"""
import torch
import argparse
import random
from sklearn import preprocessing
import yaml

try:
    from itertools import izip as zip
except ImportError:
    pass
import os
import shutil
from dance.datasets.multimodality import ModalityPredictionDataset
from dance.modules.multi_modality.predict_modality.cmae import CMAE
from dance.utils import set_seed

def prepare_sub_folder(output_directory):
    image_directory = os.path.join(output_directory, 'images')
    if not os.path.exists(image_directory):
        print("Creating directory: {}".format(image_directory))
        os.makedirs(image_directory)
    checkpoint_directory = os.path.join(output_directory, 'checkpoints')
    if not os.path.exists(checkpoint_directory):
        print("Creating directory: {}".format(checkpoint_directory))
        os.makedirs(checkpoint_directory)
    return checkpoint_directory, image_directory

def get_config(config):
    # Note need to have pip install pyyaml==5.4.1
    with open(config, 'r') as stream:
        return yaml.load(stream, Loader=yaml.Loader)

if __name__ == '__main__':
    rndseed = random.randint(0, 2147483647)
    parser = argparse.ArgumentParser()
    parser.add_argument('--config', type=str, default = './predict_modality/cross_modal_supervised.yaml', help='Path to the config file.')
    parser.add_argument('--output_path', type=str, default='./predict_modality/output', help="outputs path")
    parser.add_argument("--resume", action="store_true")
    parser.add_argument('-t', '--subtask', default='openproblems_bmmc_cite_phase2_rna')
    parser.add_argument('-device', '--device', default='cuda:3')
    parser.add_argument('-cpu', '--cpus', default = 1, type = int)
    parser.add_argument('-seed', '--rnd_seed', default = rndseed, type=int)
    opts = parser.parse_args()

    torch.set_num_threads(opts.cpus)
    rndseed = opts.rnd_seed
    set_seed(rndseed)
    dataset = ModalityPredictionDataset(opts.subtask).load_data().preprocess('feature_selection')
    device = opts.device

    # Load experiment setting
    config = get_config(opts.config)

    # Setup logger and output folders
    model_name = os.path.splitext(os.path.basename(opts.config))[0]
    output_directory = os.path.join(opts.output_path + "/outputs", model_name)
    checkpoint_directory, image_directory = prepare_sub_folder(output_directory)
    shutil.copy(opts.config, os.path.join(output_directory, 'config.yaml'))  # copy config file to output folder

    le = preprocessing.LabelEncoder()
    train_batch = dataset.modalities[0].obs['batch']
    batch = le.fit_transform(train_batch)
    train_batch = torch.from_numpy(batch).long().to(device)
    train_mod1 = torch.from_numpy(dataset.numpy_features(0)).float().to(device)
    train_mod2 = torch.from_numpy(dataset.numpy_features(1)).float().to(device)
    test_mod1 = torch.from_numpy(dataset.numpy_features(2)).float().to(device)
    test_mod2 = torch.from_numpy(dataset.numpy_features(3)).float().to(device)

    config['input_dim_a'] = dataset.modalities[0].shape[1]
    config['input_dim_b'] = dataset.modalities[1].shape[1]
    config['resume'] = opts.resume
    config['num_of_classes'] = max(batch) + 1

    model = CMAE(config)
    model.to(device)

    model.fit(train_mod1, train_mod2, train_batch, checkpoint_directory)
    print(model.predict(test_mod1))
    print(model.score(test_mod1, test_mod2))



"""
This is the implementation of scGNN.
Extended from https://github.com/juexinwang/scGNN

Reference
----------
Wang, Juexin, et al. "scGNN is a novel graph neural network framework for single-cell RNA-Seq analyses." Nature communications 12.1 (2021): 1-11.
"""

import time
import os
import sys
import igraph
from pathlib import Path
import torch
from torch import nn, optim
from torch.nn import functional as F
from torch.nn.modules.module import Module
from torch.utils.data import Dataset, DataLoader
from torch.nn.parameter import Parameter
from dance.transforms.graph_construct import scGNNgenerateAdj
import scipy.sparse as sp
import numpy as np
from tqdm import tqdm
import pickle as pkl
from sklearn.metrics import adjusted_rand_score, roc_auc_score, \
    average_precision_score
from sklearn.cluster import KMeans, SpectralClustering, AffinityPropagation, AgglomerativeClustering, \
    Birch,  MeanShift, OPTICS
import networkx as nx
import pandas as pd
import scanpy as sc
import anndata as ad


class ScDataset(Dataset):
    def __init__(self, data=None, transform=None):
        """
        Args:
            data : sparse matrix.
            transform (callable, optional):
        """
        # Now lines are cells, and cols are genes
        # self.features = data.transpose()
        self.features = data
        # save nonzero
        # self.nz_i,self.nz_j = self.features.nonzero()
        self.transform = transform

    def __len__(self):
        return self.features.shape[0]

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()

        sample = self.features[idx, :]
        if type(sample) == sp.lil_matrix:
            sample = torch.from_numpy(sample.toarray())
        else:
            sample = torch.from_numpy(sample)

        # transform after get the data
        if self.transform:
            sample = self.transform(sample)

        return sample, idx


class ScDatasetInter(Dataset):
    def __init__(self, features, transform=None):
        """
        Internal scData
        Args:
            construct dataset from features
        """
        self.features = features
        # Now lines are cells, and cols are genes
        # self.features = self.features.transpose()
        self.transform = transform

    def __len__(self):
        return self.features.shape[0]

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()

        sample = self.features[idx, :]

        # transform after get the data
        if self.transform:
            sample = self.transform(sample)

        return sample, idx


class AE(nn.Module):
    ''' Autoencoder for dimensional reduction'''

    def __init__(self, dim):
        super(AE, self).__init__()
        self.dim = dim
        self.fc1 = nn.Linear(dim, 512)
        self.fc2 = nn.Linear(512, 128)
        self.fc3 = nn.Linear(128, 512)
        self.fc4 = nn.Linear(512, dim)

    def encode(self, x):
        h1 = F.relu(self.fc1(x))
        return F.relu(self.fc2(h1))

    def decode(self, z):
        h3 = F.relu(self.fc3(z))
        return torch.relu(self.fc4(h3))

    def forward(self, x):
        z = self.encode(x)
        return self.decode(z), z


class VAE(nn.Module):
    ''' Variational Autoencoder for dimensional reduction'''

    def __init__(self, dim):
        super(VAE, self).__init__()
        self.dim = dim
        self.fc1 = nn.Linear(dim, 400)
        self.fc21 = nn.Linear(400, 20)
        self.fc22 = nn.Linear(400, 20)
        self.fc3 = nn.Linear(20, 400)
        self.fc4 = nn.Linear(400, dim)

    def encode(self, x):
        h1 = F.relu(self.fc1(x))
        return self.fc21(h1), self.fc22(h1)

    def reparameterize(self, mu, logvar):
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(std)
        return mu + eps * std

    def decode(self, z):
        h3 = F.relu(self.fc3(z))
        return torch.sigmoid(self.fc4(h3))

    def forward(self, x):
        mu, logvar = self.encode(x.view(-1, self.dim))
        z = self.reparameterize(mu, logvar)
        return self.decode(z), mu, logvar, z


class GraphConvolution(Module):
    """
    Simple GCN layer, similar to https://arxiv.org/abs/1609.02907
    """

    def __init__(self, in_features, out_features, dropout=0., act=F.relu):
        super(GraphConvolution, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        # TODO
        self.dropout = dropout
        # self.dropout = Parameter(torch.FloatTensor(dropout))
        self.act = act
        # self.weight = Parameter(torch.DoubleTensor(in_features, out_features))
        self.weight = Parameter(torch.FloatTensor(in_features, out_features))
        self.reset_parameters()

    def reset_parameters(self):
        torch.nn.init.xavier_uniform_(self.weight)

    def forward(self, input, adj):
        input = F.dropout(input, self.dropout, self.training)
        support = torch.mm(input, self.weight)
        output = torch.spmm(adj, support)
        output = self.act(output)
        return output

    def __repr__(self):
        return self.__class__.__name__ + ' (' \
               + str(self.in_features) + ' -> ' \
               + str(self.out_features) + ')'


class InnerProductDecoder(nn.Module):
    """Decoder for using inner product for prediction."""

    def __init__(self, dropout, act=torch.sigmoid):
        super(InnerProductDecoder, self).__init__()
        self.dropout = dropout
        self.act = act

    def forward(self, z):
        z = F.dropout(z, self.dropout, training=self.training)
        adj = self.act(torch.mm(z, z.t()))
        return adj


class GCNModelVAE(nn.Module):
    """
    Parameters
    ----------

    """

    def __init__(self, input_feat_dim, hidden_dim1, hidden_dim2, dropout):
        super(GCNModelVAE, self).__init__()
        self.gc1 = GraphConvolution(input_feat_dim, hidden_dim1, dropout, act=F.relu)
        self.gc2 = GraphConvolution(hidden_dim1, hidden_dim2, dropout, act=lambda x: x)
        self.gc3 = GraphConvolution(hidden_dim1, hidden_dim2, dropout, act=lambda x: x)
        self.dc = InnerProductDecoder(dropout, act=lambda x: x)

    def encode(self, x, adj):
        hidden1 = self.gc1(x, adj)
        return self.gc2(hidden1, adj), self.gc3(hidden1, adj)

    def reparameterize(self, mu, logvar):
        if self.training:
            std = torch.exp(logvar)
            eps = torch.randn_like(std)
            return eps.mul(std).add_(mu)
        else:
            return mu

    def forward(self, x, adj):
        mu, logvar = self.encode(x, adj)
        z = self.reparameterize(mu, logvar)
        return z, mu, logvar


class GCNModelAE(nn.Module):
    """
    Parameters
    ----------

    """

    def __init__(self, input_feat_dim, hidden_dim1, hidden_dim2, dropout):
        super(GCNModelAE, self).__init__()
        self.gc1 = GraphConvolution(input_feat_dim, hidden_dim1, dropout, act=F.relu)
        self.gc2 = GraphConvolution(hidden_dim1, hidden_dim2, dropout, act=lambda x: x)
        self.dc = InnerProductDecoder(dropout, act=lambda x: x)

    def encode(self, x, adj):
        hidden1 = self.gc1(x, adj)
        return self.gc2(hidden1, adj)

    def forward(self, x, adj, encode=False):
        z = self.encode(x, adj)
        return z, z, None


class scGNN():
    """ Single-cell GNN model
    Parameters
    ----------
    adata :
        total data in anndata format
    genelist :
        list of gene names in data
    celllist :
        list of cell names in data
    params :
        model parameters
    model_type : str
        AE or VAE, specifying model architecture
    n_hidden2 : int
        specifies hidden dimension
    activation :
        gives torch activation function for model
    dropout : float
        probability of weight dropout during training
    GAEepochs : int
        number of epochs for GAE
    Regu_epochs : int
        number of epochs for main regulizing model in cycle
    EM_epochs : int
        number of Expectation Maximation cycles
    cluster_epochs : int
        number of epochs for cluster encoders
    batch_size : int
        number of samples in minibatch
    debugMode : str
        alternative training with debug
    """

    def __init__(self, adata, genelist, celllist, params, model_type='AE', n_hidden2=32, activation=F.relu, dropout=0.1,
                 GAEepochs=200, Regu_epochs=500, EM_epochs=200, cluster_epochs=200, batch_size=128,
                 debugMode='noDebug'):  # , lr = 1e-3, weight_decay = .9):
        self.params = params
        self.postfix = time.strftime('%d_%m_%Y') + '_' + time.strftime('%H:%M:%S')
        self.prj_path = Path(
            __file__).parent.resolve().parent.resolve().parent.resolve().parent.resolve().parent.resolve()
        self.save_path = self.prj_path / 'example' / 'single_modality' / 'imputation' / 'pretrained' / params.train_dataset / 'models'
        if not self.save_path.exists():
            self.save_path.mkdir(parents=True)
        self.device = params.device
        self.GAEepochs = GAEepochs
        self.Regu_epochs = Regu_epochs
        self.EM_epochs = EM_epochs
        self.cluster_epochs = cluster_epochs
        self.debugMode = debugMode
        # self.lr = lr
        self.batch_size = batch_size
        # self.n_hidden1 = n_hidden1
        # self.n_hidden2 = n_hidden2
        self.activation = activation
        self.adata = adata
        self.num_cells = self.adata.shape[0]
        self.num_genes = self.adata.shape[1]
        self.genelist = genelist
        self.celllist = celllist
        # self.train_data = train_data
        # self.test_data = self.dl_params.test_data.toarray()
        self.model_type = model_type
        # self.train_ids = torch.Tensor(self.train_ids.toarray())
        # self.test_ids = torch.Tensor(self.test_ids.toarray())
        # if self.model_type == 'AE':
        #     self.model = AE(dim=self.num_genes).to(self.device)
        # else:
        #     self.model = VAE(dim=self.num_genes).to(self.device)
        # self.optimizer = torch.optim.Adam(self.model.parameters(), lr=1e-3)
        # weight_decay=self.weight_decay)
        self.kwargs = {'num_workers': 0, 'pin_memory': False} if params.cuda else {}

    def get_enum(self, reduction):
        """ Get enumeration helper function
        Parameters
        ----------
        reduction : str
            none, mean, elementwise_mean, sum
        Returns
        ----------
        ret : int
        """
        if reduction == 'none':
            ret = 0
        elif reduction == 'mean':
            ret = 1
        elif reduction == 'elementwise_mean':
            print("reduction='elementwise_mean' is deprecated, please use reduction='mean' instead.")
            ret = 1
        elif reduction == 'sum':
            ret = 2
        else:
            ret = -1  # remove once JIT exceptions support control flow
            raise ValueError(
                "{} is not a valid value for reduction".format(reduction))
        return ret

    def generateCelltypeRegu(self, listResult):
        """
        Parameters
        ----------
        listResult :
            list of cell types
        Returns
        ----------
        celltypesample :
            matrix of cell types
        """
        celltypesample = np.zeros((len(listResult), len(listResult)))
        tdict = {}
        count = 0
        for item in listResult:
            if item in tdict:
                tlist = tdict[item]
            else:
                tlist = []
            tlist.append(count)
            tdict[item] = tlist
            count += 1

        for key in sorted(tdict):
            tlist = tdict[key]
            for item1 in tlist:
                for item2 in tlist:
                    celltypesample[item1, item2] = 1.0

        return celltypesample

    def sample_mask(self, idx, l):
        """Create mask.
        Parameters
        ----------
        idx :
            boolean matrix for coordinates
        l :
            size of mask
        Returns
        ----------
        mask :
            Boolean matrix for mask
        """
        mask = np.zeros(l)
        mask[idx] = 1
        mask = np.array(mask, dtype=np.bool)
        return mask

    def legacy_get_string(self, size_average, reduce, emit_warning=True):
        warning = "size_average and reduce args will be deprecated, please use reduction='{}' instead."

        if size_average is None:
            size_average = True
        if reduce is None:
            reduce = True

        if size_average and reduce:
            ret = 'mean'
        elif reduce:
            ret = 'sum'
        else:
            ret = 'none'
        if emit_warning:
            print(warning.format(ret))
        return ret

    def vallina_mse_loss_function(self, input, target, size_average=None, reduce=None, reduction='mean'):
        r"""vallina_mse_loss_function(input, target, size_average=None, reduce=None, reduction='mean') -> Tensor
        Original: Measures the element-wise mean squared error.
        See :revised from pytorch class:`~torch.nn.MSELoss` for details.
        """
        if not (target.size() == input.size()):
            print("Using a target size ({}) that is different to the input size ({}). "
                  "This will likely lead to incorrect results due to broadcasting. "
                  "Please ensure they have the same size.".format(target.size(), input.size()))
        if size_average is not None or reduce is not None:
            reduction = self.legacy_get_string(size_average, reduce)
        # Now it use regulariz type to distinguish, it can be imporved later
        # Original, for not require grads, using c++ version
        # However, it has bugs there, different number of cpu cause different results because of MKL parallel library
        # Not known yet whether GPU has same problem.
        # Solution 1: set same number of cpu when running, it works for reproduce everything but not applicable for other users
        # https://pytorch.org/docs/stable/torch.html#torch.set_num_threads
        # https://pytorch.org/docs/stable/notes/cpu_threading_torchscript_inference.html
        # Solution 2: not use C++ codes, as we did here.
        # https://github.com/pytorch/pytorch/issues/8710

        if target.requires_grad:
            ret = (input - target) ** 2
            # 0.001 to reduce float loss
            # ret = (0.001*input - 0.001*target) ** 2
            if reduction != 'none':
                ret = torch.mean(ret) if reduction == 'mean' else torch.sum(ret)
        else:
            expanded_input, expanded_target = torch.broadcast_tensors(
                input, target)
            ret = F.mse_loss(
                expanded_input, expanded_target, reduction=reduction)

        # ret = (input - target) ** 2
        # # 0.001 to reduce float loss
        # # ret = (0.001*input - 0.001*target) ** 2
        # if reduction != 'none':
        #     ret = torch.mean(ret) if reduction == 'mean' else torch.sum(ret)
        return ret

    def regulation_mse_loss_function(self, input, target, regulationMatrix, size_average=None, reduce=None,
                                     reduction='mean'):
        r"""regulation_mse_loss_function(input, target, regulationMatrix, regularizer_type, size_average=None, reduce=None, reduction='mean') -> Tensor
        Measures the element-wise mean squared error for regulation input, now only support LTMG.
        See :revised from pytorch class:`~torch.nn.MSELoss` for details.
        """
        if not (target.size() == input.size()):
            print("Using a target size ({}) that is different to the input size ({}). "
                  "This will likely lead to incorrect results due to broadcasting. "
                  "Please ensure they have the same size.".format(target.size(), input.size()))
        if size_average is not None or reduce is not None:
            reduction = self.legacy_get_string(size_average, reduce)
        # Now it use regulariz type to distinguish, it can be imporved later
        ret = (input - target) ** 2
        # ret = (0.001*input - 0.001*target) ** 2
        ret = torch.mul(ret, regulationMatrix)
        if reduction != 'none':
            ret = torch.mean(ret) if reduction == 'mean' else torch.sum(ret)
        return ret

    def graph_mse_loss_function(self, input, target, graphregu, size_average=None, reduce=None, reduction='mean'):
        r"""graph_mse_loss_function(input, target, adj, regularizer_type, size_average=None, reduce=None, reduction='mean') -> Tensor
        Measures the element-wise mean squared error in graph regularizor.
        See:revised from pytorch class:`~torch.nn.MSELoss` for details.
        """
        if not (target.size() == input.size()):
            print("Using a target size ({}) that is different to the input size ({}). "
                  "This will likely lead to incorrect results due to broadcasting. "
                  "Please ensure they have the same size.".format(target.size(), input.size()))
        if size_average is not None or reduce is not None:
            reduction = self.legacy_get_string(size_average, reduce)
        # Now it use regulariz type to distinguish, it can be imporved later
        ret = (input - target) ** 2
        # ret = (0.001*input - 0.001*target) ** 2
        # if graphregu != None:
        # print(graphregu.type())
        # print(ret.type())
        ret = torch.matmul(graphregu, ret)
        if reduction != 'none':
            ret = torch.mean(ret) if reduction == 'mean' else torch.sum(ret)
        return ret

    def regulation01_mse_loss_function(self, input, target, regulationMatrix, size_average=None, reduce=None,
                                       reduction='mean'):
        r"""regulation_mse_loss_function(input, target, regulationMatrix, regularizer_type, size_average=None, reduce=None, reduction='mean') -> Tensor
        Measures the element-wise mean squared error for regulation input, now only support LTMG.
        See :revised from pytorch class:`~torch.nn.MSELoss` for details.
        """
        if not (target.size() == input.size()):
            print("Using a target size ({}) that is different to the input size ({}). "
                  "This will likely lead to incorrect results due to broadcasting. "
                  "Please ensure they have the same size.".format(target.size(), input.size()))
        if size_average is not None or reduce is not None:
            reduction = self.legacy_get_string(size_average, reduce)
        # Now it use regulariz type to distinguish, it can be imporved later
        ret = (input - target) ** 2
        # ret = (0.001*input - 0.001*target) ** 2
        regulationMatrix[regulationMatrix > 0] = 1
        ret = torch.mul(ret, regulationMatrix)
        if reduction != 'none':
            ret = torch.mean(ret) if reduction == 'mean' else torch.sum(ret)
        return ret

    def loss_function_graph(self, recon_x, x, mu, logvar, graphregu=None, gammaPara=1.0, regulationMatrix=None,
                            regularizer_type='noregu', reguPara=0.001, modelusage='AE', reduction='sum'):
        '''
        Regularized by the graph information
        Reconstruction + KL divergence losses summed over all elements and batch
        '''
        # Original
        # BCE = F.binary_cross_entropy(recon_x, x.view(-1, 784), reduction='sum')
        # Graph
        target = x
        if regularizer_type == 'Graph' or regularizer_type == 'LTMG' or regularizer_type == 'LTMG01':
            target.requires_grad = True
        else:
            target = x.detach()

        # Euclidean
        # BCE = gammaPara * vallina_mse_loss_function(recon_x, target, reduction='sum')
        BCE = gammaPara * \
              self.vallina_mse_loss_function(recon_x, target, reduction=reduction)
        if regularizer_type == 'noregu':
            loss = BCE
        elif regularizer_type == 'LTMG':
            loss = BCE + reguPara * \
                   self.regulation_mse_loss_function(
                       recon_x, target, regulationMatrix, reduction=reduction)
        elif regularizer_type == 'LTMG01':
            loss = BCE + reguPara * \
                   self.regulation01_mse_loss_function(
                       recon_x, target, regulationMatrix, reduction=reduction)
        elif regularizer_type == 'Graph':
            loss = BCE + reguPara * \
                   self.graph_mse_loss_function(
                       recon_x, target, graphregu=graphregu, reduction=reduction)
        elif regularizer_type == 'GraphR':
            loss = BCE + reguPara * \
                   self.graph_mse_loss_function(
                       recon_x, target, graphregu=1 - graphregu, reduction=reduction)
        elif regularizer_type == 'LTMG-Graph':
            loss = BCE + reguPara * self.regulation_mse_loss_function(recon_x, target, regulationMatrix,
                                                                      reduction=reduction) + \
                   reguPara * \
                   self.graph_mse_loss_function(
                       recon_x, target, graphregu=graphregu, reduction=reduction)
        elif regularizer_type == 'LTMG-GraphR':
            loss = BCE + reguPara * self.regulation_mse_loss_function(recon_x, target, regulationMatrix,
                                                                      reduction=reduction) + \
                   reguPara * \
                   self.graph_mse_loss_function(
                       recon_x, target, graphregu=1 - graphregu, reduction=reduction)

        # see Appendix B from VAE paper:
        # Kingma and Welling. Auto-Encoding Variational Bayes. ICLR, 2014
        # https://arxiv.org/abs/1312.6114
        # 0.5 * sum(1 + log(sigma^2) - mu^2 - sigma^2)
        if modelusage == 'VAE':
            KLD = -0.5 * torch.sum(1 + logvar - mu.pow(2) - logvar.exp())
            loss = loss + KLD

        return loss

    def loss_function_graph_celltype(self, recon_x, x, mu, logvar, graphregu=None, celltyperegu=None, gammaPara=1.0,
                                     regulationMatrix=None, regularizer_type='noregu', reguPara=0.001,
                                     reguParaCelltype=0.001, modelusage='AE', reduction='sum'):
        '''
        Regularized by the graph information
        Reconstruction + KL divergence losses summed over all elements and batch
        '''
        # Original
        # BCE = F.binary_cross_entropy(recon_x, x.view(-1, 784), reduction='sum')
        # Graph
        target = x
        # if regularizer_type == 'Graph' or regularizer_type == 'LTMG' or regularizer_type == 'LTMG01' or regularizer_type == 'Celltype':
        #    target.requires_grad = True
        # Euclidean
        # BCE = gammaPara * vallina_mse_loss_function(recon_x, target, reduction='sum')
        BCE = gammaPara * \
              self.vallina_mse_loss_function(recon_x, target, reduction=reduction)
        if regularizer_type == 'noregu':
            loss = BCE
        elif regularizer_type == 'LTMG':
            loss = BCE + reguPara * \
                   self.regulation_mse_loss_function(
                       recon_x, target, regulationMatrix, reduction=reduction)
        elif regularizer_type == 'LTMG01':
            loss = BCE + reguPara * \
                   self.regulation01_mse_loss_function(
                       recon_x, target, regulationMatrix, reduction=reduction)
        elif regularizer_type == 'Graph':
            loss = BCE + reguPara * \
                   self.graph_mse_loss_function(
                       recon_x, target, graphregu=graphregu, reduction=reduction)
        elif regularizer_type == 'Celltype':
            loss = BCE + reguPara * self.graph_mse_loss_function(recon_x, target, graphregu=graphregu,
                                                                 reduction=reduction) + \
                   reguParaCelltype * \
                   self.graph_mse_loss_function(
                       recon_x, target, graphregu=celltyperegu, reduction=reduction)
        elif regularizer_type == 'CelltypeR':
            loss = BCE + (1 - gammaPara) * self.regulation01_mse_loss_function(recon_x, target, regulationMatrix,
                                                                               reduction=reduction) + reguPara * self.graph_mse_loss_function(
                recon_x,
                target, graphregu=graphregu, reduction=reduction) + reguParaCelltype * self.graph_mse_loss_function(
                recon_x,
                target,
                graphregu=celltyperegu,
                reduction=reduction)

        # see Appendix B from VAE paper:
        # Kingma and Welling. Auto-Encoding Variational Bayes. ICLR, 2014
        # https://arxiv.org/abs/1312.6114
        # 0.5 * sum(1 + log(sigma^2) - mu^2 - sigma^2)
        if modelusage == 'VAE':
            KLD = -0.5 * torch.sum(1 + logvar - mu.pow(2) - logvar.exp())
            loss = loss + KLD

        return loss

    def gae_loss_function(self, preds, labels, mu, logvar, n_nodes, norm, pos_weight):
        """ Loss function for Graph AutoEncoder
        Parameters
        ----------
        preds :
            prediction matrix
        labels :
            true matrix
        mu : float
            parameter in KL distance
        logvar : float
            log variance
        n_nodes : int
            number of nodes in graph
        norm : float
            adjacency norm
        pos_weight : float
            weight of positive examples in cross entropy

        Returns
        ----------
        loss : float
            loss summation of cross entropy and KL distance
        """
        cost = norm * F.binary_cross_entropy_with_logits(preds, labels, pos_weight=labels * pos_weight)

        # Check if the model is simple Graph Auto-encoder
        if logvar is None:
            return cost

        # see Appendix B from VAE paper:
        # Kingma and Welling. Auto-Encoding Variational Bayes. ICLR, 2014
        # https://arxiv.org/abs/1312.6114
        # 0.5 * sum(1 + log(sigma^2) - mu^2 - sigma^2)
        KLD = -0.5 / n_nodes * torch.mean(torch.sum(
            1 + 2 * logvar - mu.pow(2) - logvar.exp().pow(2), 1))
        loss = cost + KLD
        return loss

    def trimClustering(self, listResult, minMemberinCluster=5, maxClusterNumber=30):
        '''
        If the clustering numbers larger than certain number, use this function to trim.
        '''
        numDict = {}
        for item in listResult:
            if not item in numDict:
                numDict[item] = 0
            else:
                numDict[item] = numDict[item] + 1

        size = len(set(listResult))
        changeDict = {}
        for item in range(size):
            if numDict[item] < minMemberinCluster or item >= maxClusterNumber:
                changeDict[item] = ''

        count = 0
        for item in listResult:
            if item in changeDict:
                listResult[count] = maxClusterNumber
            count += 1

        return listResult

    def generateLouvainCluster(self, edgeList):
        """
        Louvain Clustering using igraph
        """
        Gtmp = nx.Graph()
        Gtmp.add_weighted_edges_from(edgeList)
        W = nx.adjacency_matrix(Gtmp)
        W = W.todense()
        graph = igraph.Graph.Weighted_Adjacency(
            W.tolist(), mode=igraph.ADJ_UNDIRECTED, attr="weight", loops=False)
        louvain_partition = graph.community_multilevel(
            weights=graph.es['weight'], return_levels=False)
        size = len(louvain_partition)
        hdict = {}
        count = 0
        for i in range(size):
            tlist = louvain_partition[i]
            for j in range(len(tlist)):
                hdict[tlist[j]] = i
                count += 1

        listResult = []
        for i in range(count):
            listResult.append(hdict[i])

        return listResult, size

    def sparse_to_tuple(self, sparse_mx):
        if not sp.isspmatrix_coo(sparse_mx):
            sparse_mx = sparse_mx.tocoo()
        coords = np.vstack((sparse_mx.row, sparse_mx.col)).transpose()
        values = sparse_mx.data
        shape = sparse_mx.shape
        return coords, values, shape

    def mask_test_edges(self, adj):
        """ Function to build test set with 10% positive links
        Parameters
        -------------
        adj :
            adjacency matrix
        Returns
        -------------

        """
        #
        # NOTE: Splits are randomized and results might slightly deviate from reported numbers in the paper.
        # TODO: Clean up.

        # Remove diagonal elements
        adj = adj - sp.dia_matrix((adj.diagonal()[np.newaxis, :], [0]), shape=adj.shape)
        adj.eliminate_zeros()
        # Check that diag is zero:
        assert np.diag(adj.todense()).sum() == 0

        adj_triu = sp.triu(adj)
        adj_tuple = self.sparse_to_tuple(adj_triu)
        edges = adj_tuple[0]
        edges_all = self.sparse_to_tuple(adj)[0]
        num_test = int(np.floor(edges.shape[0] / 10.))
        num_val = int(np.floor(edges.shape[0] / 20.))

        all_edge_idx = np.arange(edges.shape[0])
        np.random.shuffle(all_edge_idx)
        val_edge_idx = all_edge_idx[:num_val]
        test_edge_idx = all_edge_idx[num_val:(num_val + num_test)]
        test_edges = edges[test_edge_idx]
        val_edges = edges[val_edge_idx]
        train_edges = np.delete(edges, np.hstack([test_edge_idx, val_edge_idx]), axis=0)

        def ismember(a, b, tol=5):
            rows_close = np.all(np.round(a - b[:, None], tol) == 0, axis=-1)
            return np.any(rows_close)

        test_edges_false = []
        while len(test_edges_false) < len(test_edges):
            idx_i = np.random.randint(0, adj.shape[0])
            idx_j = np.random.randint(0, adj.shape[0])
            if idx_i == idx_j:
                continue
            if ismember([idx_i, idx_j], edges_all):
                continue
            if test_edges_false:
                if ismember([idx_j, idx_i], np.array(test_edges_false)):
                    continue
                if ismember([idx_i, idx_j], np.array(test_edges_false)):
                    continue
            test_edges_false.append([idx_i, idx_j])

        val_edges_false = []
        while len(val_edges_false) < len(val_edges):
            idx_i = np.random.randint(0, adj.shape[0])
            idx_j = np.random.randint(0, adj.shape[0])
            if idx_i == idx_j:
                continue
            if ismember([idx_i, idx_j], train_edges):
                continue
            if ismember([idx_j, idx_i], train_edges):
                continue
            if ismember([idx_i, idx_j], val_edges):
                continue
            if ismember([idx_j, idx_i], val_edges):
                continue
            if val_edges_false:
                if ismember([idx_j, idx_i], np.array(val_edges_false)):
                    continue
                if ismember([idx_i, idx_j], np.array(val_edges_false)):
                    continue
            if ~ismember([idx_i, idx_j], edges_all) and ~ismember([idx_j, idx_i], edges_all):
                val_edges_false.append([idx_i, idx_j])
            else:
                # Debug
                print(str(idx_i) + " " + str(idx_j))
            # Original:
            # val_edges_false.append([idx_i, idx_j])

        # TODO: temporary disable for ismember function may require huge memory.
        # assert ~ismember(test_edges_false, edges_all)
        # assert ~ismember(val_edges_false, edges_all)
        # assert ~ismember(val_edges, train_edges)
        # assert ~ismember(test_edges, train_edges)
        # assert ~ismember(val_edges, test_edges)

        data = np.ones(train_edges.shape[0])

        # Re-build adj matrix
        adj_train = sp.csr_matrix((data, (train_edges[:, 0], train_edges[:, 1])), shape=adj.shape)
        adj_train = adj_train + adj_train.T

        # NOTE: these edge lists only contain single direction of edge!
        return adj_train, train_edges, val_edges, val_edges_false, test_edges, test_edges_false

    def sparse_mx_to_torch_sparse_tensor(self, sparse_mx):
        """Convert a scipy sparse matrix to a torch sparse tensor."""
        # sparse_mx = sparse_mx.tocoo().astype(np.float64)
        sparse_mx = sparse_mx.tocoo().astype(np.float32)
        indices = torch.from_numpy(
            np.vstack((sparse_mx.row, sparse_mx.col)).astype(np.int64))
        values = torch.from_numpy(sparse_mx.data)
        shape = torch.Size(sparse_mx.shape)
        # return torch.sparse.DoubleTensor(indices, values, shape)
        return torch.sparse.FloatTensor(indices, values, shape)

    def preprocess_graph(self, adj):
        adj = sp.coo_matrix(adj)
        adj_ = adj + sp.eye(adj.shape[0])
        rowsum = np.array(adj_.sum(1))
        degree_mat_inv_sqrt = sp.diags(np.power(rowsum, -0.5).flatten())
        adj_normalized = adj_.dot(degree_mat_inv_sqrt).transpose().dot(degree_mat_inv_sqrt).tocoo()
        # return sparse_to_tuple(adj_normalized)
        return self.sparse_mx_to_torch_sparse_tensor(adj_normalized)

    def get_roc_score(self, emb, adj_orig, edges_pos, edges_neg):
        def sigmoid(x):
            return 1 / (1 + np.exp(-x))

        # Predict on test set of edges
        adj_rec = np.dot(emb, emb.T)
        preds = []
        pos = []
        for e in edges_pos:
            preds.append(sigmoid(adj_rec[e[0], e[1]]))
            pos.append(adj_orig[e[0], e[1]])

        preds_neg = []
        neg = []
        for e in edges_neg:
            preds_neg.append(sigmoid(adj_rec[e[0], e[1]]))
            neg.append(adj_orig[e[0], e[1]])

        preds_all = np.hstack([preds, preds_neg])
        labels_all = np.hstack([np.ones(len(preds)), np.zeros(len(preds))])
        roc_score = roc_auc_score(labels_all, preds_all)
        ap_score = average_precision_score(labels_all, preds_all)

        return roc_score, ap_score

    def GAEembedding(self, z, adj, params):
        '''
        GAE embedding for clustering
        Parameters
        ----------
        z :
            input parameters
        adj :
            adjacency matrix of cell graph
        params :
            model parameters
        Returns
        ----------
        hidden_emb :
            Embedding from graph
        '''
        # featrues from z
        # Louvain
        features = z
        # features = torch.DoubleTensor(features)
        features = torch.FloatTensor(features)

        # Old implementation
        # adj, features, y_test, tx, ty, test_maks, true_labels = load_data(params.dataset_str)

        n_nodes, feat_dim = features.shape

        # Store original adjacency matrix (without diagonal entries) for later
        adj_orig = adj
        adj_orig = adj_orig - sp.dia_matrix((adj_orig.diagonal()[np.newaxis, :], [0]), shape=adj_orig.shape)
        adj_orig.eliminate_zeros()

        adj_train, train_edges, val_edges, val_edges_false, test_edges, test_edges_false = self.mask_test_edges(adj)
        adj = adj_train

        # Some preprocessing
        adj_norm = self.preprocess_graph(adj)
        adj_label = adj_train + sp.eye(adj_train.shape[0])
        # adj_label = sparse_to_tuple(adj_label)
        # adj_label = torch.DoubleTensor(adj_label.toarray())
        adj_label = torch.FloatTensor(adj_label.toarray())

        pos_weight = float(adj.shape[0] * adj.shape[0] - adj.sum()) / adj.sum()
        norm = adj.shape[0] * adj.shape[0] / float((adj.shape[0] * adj.shape[0] - adj.sum()) * 2)

        if params.GAEmodel == 'gcn_vae':
            model = GCNModelVAE(feat_dim, params.GAEhidden1, params.GAEhidden2, params.GAEdropout)
        else:
            model = GCNModelAE(feat_dim, params.GAEhidden1, params.GAEhidden2, params.GAEdropout)
        if params.precisionModel == 'Double':
            model = model.double()
        optimizer = optim.Adam(model.parameters(), lr=params.GAElr)

        for epoch in tqdm(range(params.GAEepochs)):
            t = time.time()
            # mem=resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
            # print('Mem consumption before training: '+str(mem))
            model.train()
            optimizer.zero_grad()
            z, mu, logvar = model(features, adj_norm)

            loss = self.gae_loss_function(preds=model.dc(z), labels=adj_label,
                                          mu=mu, logvar=logvar, n_nodes=n_nodes,
                                          norm=norm, pos_weight=pos_weight)
            loss.backward()
            cur_loss = loss.item()
            optimizer.step()

            hidden_emb = mu.data.numpy()

            # roc_curr, ap_curr = get_roc_score(hidden_emb, adj_orig, val_edges, val_edges_false)
            ap_curr = 0

            tqdm.write("Epoch: {}, train_loss_gae={:.5f}, val_ap={:.5f}, time={:.5f}".format(
                epoch + 1, cur_loss,
                ap_curr, time.time() - t))

        tqdm.write("Optimization Finished!")

        roc_score, ap_score = self.get_roc_score(hidden_emb, adj_orig, test_edges, test_edges_false)
        tqdm.write('Test ROC score: ' + str(roc_score))
        tqdm.write('Test AP score: ' + str(ap_score))

        return hidden_emb

    def train(self, model, optimizer, epoch, params, train_loader, regulationMatrix=None, EMFlag=False,
              taskType='celltype', sparseImputation='nonsparse', ):
        ''' Training loop
        Parameters
        ----------
        model :
            input model to train
        optimizer :
            input optimizer for model training
        epoch : int
            epoch number
        params :
            model parameters
        train_loader :
            input data class
        regulationMatrix
            LTMG regulation matrix
        EMFlag :
            EMFlag indicates whether in EM processes.
            If in EM, use regulized-type parsed from program entrance,
            Otherwise, noregu
        taskType : str
            celltype or imputation for different model training
        sparseImputation : str
            whether input is sparse format
        Returns
        ----------
        recon_batch_all :
            reconstruction of batch data
        data_all :
            return input data from train_loader
        z_all :
            hidden representation of data from model
        '''
        model.train()
        train_loss = 0
        for batch_idx, (data, dataindex) in enumerate(train_loader):
            data = data.to(self.device)
            if params.precisionModel == 'Double':
                data = data.type(torch.DoubleTensor)
            elif params.precisionModel == 'Float':
                data = data.type(torch.FloatTensor)
            data = data.to(self.device)
            if not params.regulized_type == 'noregu':
                regulationMatrixBatch = regulationMatrix[dataindex, :]
                regulationMatrixBatch = regulationMatrixBatch.to(self.device)
            else:
                regulationMatrixBatch = None
            if taskType == 'imputation':
                if sparseImputation == 'nonsparse':
                    celltypesampleBatch = self.celltypesample[dataindex,
                                          :][:, dataindex]
                    adjsampleBatch = self.adjsample[dataindex, :][:, dataindex]
                elif sparseImputation == 'sparse':
                    celltypesampleBatch = self.generateCelltypeRegu(
                        self.listResult[dataindex])
                    celltypesampleBatch = torch.from_numpy(celltypesampleBatch)
                    if params.precisionModel == 'Float':
                        celltypesampleBatch = celltypesampleBatch.float()
                    elif params.precisionModel == 'Double':
                        celltypesampleBatch = celltypesampleBatch.type(
                            torch.DoubleTensor)
                    # mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
                    # print('celltype Mem consumption: '+str(mem))

                    adjsampleBatch = self.adj[dataindex, :][:, dataindex]
                    adjsampleBatch = sp.csr_matrix.todense(adjsampleBatch)
                    adjsampleBatch = torch.from_numpy(adjsampleBatch)
                    if params.precisionModel == 'Float':
                        adjsampleBatch = adjsampleBatch.float()
                    elif params.precisionModel == 'Double':
                        adjsampleBatch = adjsampleBatch.type(torch.DoubleTensor)
                    # mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
                    # print('adj Mem consumption: '+str(mem))
            optimizer.zero_grad()
            if self.model_type == 'VAE':
                recon_batch, mu, logvar, z = model(data)
                if taskType == 'celltype':
                    if EMFlag and (not params.EMreguTag):
                        loss = self.loss_function_graph(recon_batch, data.view(-1, recon_batch.shape[1]), mu, logvar,
                                                        gammaPara=params.gammaPara,
                                                        regulationMatrix=regulationMatrixBatch,
                                                        regularizer_type='noregu', reguPara=params.alphaRegularizePara,
                                                        modelusage=self.model_type, reduction=params.reduction)
                    else:
                        loss = self.loss_function_graph(recon_batch, data.view(-1, recon_batch.shape[1]), mu, logvar,
                                                        gammaPara=params.gammaPara,
                                                        regulationMatrix=regulationMatrixBatch,
                                                        regularizer_type=params.regulized_type,
                                                        reguPara=params.alphaRegularizePara, modelusage=self.model_type,
                                                        reduction=params.reduction)
                elif taskType == 'imputation':
                    if EMFlag and (not params.EMreguTag):
                        loss = self.loss_function_graph_celltype(recon_batch, data.view(-1, recon_batch.shape[1]), mu,
                                                                 logvar, graphregu=adjsampleBatch,
                                                                 celltyperegu=celltypesampleBatch,
                                                                 gammaPara=params.gammaImputePara,
                                                                 regulationMatrix=regulationMatrixBatch,
                                                                 regularizer_type=params.EMregulized_type,
                                                                 reguPara=params.graphImputePara,
                                                                 reguParaCelltype=params.celltypeImputePara,
                                                                 modelusage=self.model_type, reduction=params.reduction)
                    else:
                        loss = self.loss_function_graph_celltype(recon_batch, data.view(-1, recon_batch.shape[1]), mu,
                                                                 logvar, graphregu=adjsampleBatch,
                                                                 celltyperegu=celltypesampleBatch,
                                                                 gammaPara=params.gammaImputePara,
                                                                 regulationMatrix=regulationMatrixBatch,
                                                                 regularizer_type=params.regulized_type,
                                                                 reguPara=params.graphImputePara,
                                                                 reguParaCelltype=params.celltypeImputePara,
                                                                 modelusage=self.model_type, reduction=params.reduction)

            elif self.model_type == 'AE':
                recon_batch, z = model(data)
                mu_dummy = ''
                logvar_dummy = ''
                if taskType == 'celltype':
                    if EMFlag and (not params.EMreguTag):
                        loss = self.loss_function_graph(recon_batch, data.view(-1, recon_batch.shape[1]), mu_dummy,
                                                        logvar_dummy, gammaPara=params.gammaPara,
                                                        regulationMatrix=regulationMatrixBatch,
                                                        regularizer_type='noregu', reguPara=params.alphaRegularizePara,
                                                        modelusage=self.model_type, reduction=params.reduction)
                    else:
                        loss = self.loss_function_graph(recon_batch, data.view(-1, recon_batch.shape[1]), mu_dummy,
                                                        logvar_dummy, gammaPara=params.gammaPara,
                                                        regulationMatrix=regulationMatrixBatch,
                                                        regularizer_type=params.regulized_type,
                                                        reguPara=params.alphaRegularizePara, modelusage=self.model_type,
                                                        reduction=params.reduction)
                elif taskType == 'imputation':
                    if EMFlag and (not params.EMreguTag):
                        loss = self.loss_function_graph_celltype(recon_batch, data.view(-1, recon_batch.shape[1]),
                                                                 mu_dummy, logvar_dummy, graphregu=adjsampleBatch,
                                                                 celltyperegu=celltypesampleBatch,
                                                                 gammaPara=params.gammaImputePara,
                                                                 regulationMatrix=regulationMatrixBatch,
                                                                 regularizer_type=params.EMregulized_type,
                                                                 reguPara=params.graphImputePara,
                                                                 reguParaCelltype=params.celltypeImputePara,
                                                                 modelusage=self.model_type, reduction=params.reduction)
                    else:
                        loss = self.loss_function_graph_celltype(recon_batch, data.view(-1, recon_batch.shape[1]),
                                                                 mu_dummy, logvar_dummy, graphregu=adjsampleBatch,
                                                                 celltyperegu=celltypesampleBatch,
                                                                 gammaPara=params.gammaImputePara,
                                                                 regulationMatrix=regulationMatrixBatch,
                                                                 regularizer_type=params.regulized_type,
                                                                 reguPara=params.graphImputePara,
                                                                 reguParaCelltype=params.celltypeImputePara,
                                                                 modelusage=self.model_type, reduction=params.reduction)

            # L1 and L2 regularization
            # 0.0 for no regularization
            l1 = 0.0
            l2 = 0.0
            for p in model.parameters():
                l1 = l1 + p.abs().sum()
                l2 = l2 + p.pow(2).sum()
            # loss = loss + params.L1Para * l1 + params.L2Para * l2
            train_loss += loss.item()
            loss.backward()
            optimizer.step()
            if batch_idx % params.log_interval == 0:
                print('Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}'.format(
                    epoch, batch_idx * len(data), len(train_loader.dataset),
                           100. * batch_idx / len(train_loader),
                           train_loss / len(data)))

            # for batch
            if batch_idx == 0:
                recon_batch_all = recon_batch
                data_all = data
                z_all = z
            else:
                recon_batch_all = torch.cat((recon_batch_all, recon_batch), 0)
                data_all = torch.cat((data_all, data), 0)
                z_all = torch.cat((z_all, z), 0)

        print('====> Epoch: {} Average loss: {:.4f}'.format(
            epoch, train_loss / len(train_loader.dataset)))

        return recon_batch_all, data_all, z_all

    def fit(self, train_data):
        """ scGNN fit function
        Parameters
        ----------
        train_data :
            input gene expression features
        Returns
        ----------
        reconOri :
            Initial data reconstruction used for final imputation
        """
        if sp.issparse(train_data):
            train_data = train_data.toarray()
        data = train_data
        scData = ScDataset(data)
        train_loader = DataLoader(
            scData, batch_size=self.batch_size, shuffle=False, **self.kwargs)
        print('---TrainLoader has been successfully prepared.')
        if self.model_type == 'VAE':
            model = VAE(dim=data.shape[1]).to(self.device)
        elif self.model_type == 'AE':
            model = AE(dim=data.shape[1]).to(self.device)
        if self.params.precisionModel == 'Double':
            model = model.double()
        optimizer = torch.optim.Adam(model.parameters(), lr=1e-3)
        print('---Pytorch model ready.')
        if not os.path.exists(self.params.outputDir):
            os.makedirs(self.params.outputDir)
        ptfileStart = self.params.outputDir + self.params.train_dataset + '_EMtrainingStart.pt'

        if self.params.debugMode == 'savePrune' or self.params.debugMode == 'noDebug':
            stateStart = {
                # 'epoch': epoch,
                'state_dict': model.state_dict(),
                'optimizer': optimizer.state_dict(),
            }
            torch.save(stateStart, ptfileStart)
            for epoch in range(1, self.Regu_epochs + 1):
                recon, original, z = self.train(model, optimizer, epoch, self.params, train_loader, EMFlag=False)

            zOut = z.detach().cpu().numpy()
            print('zOut ready')
            ptstatus = model.state_dict()

            # Store reconOri for imputation
            reconOri = recon.clone()
            reconOri = reconOri.detach().cpu().numpy()

            # Step 1. Inferring celltype

            self.adj, self.edgeList = scGNNgenerateAdj(zOut, graphType=self.params.prunetype,
                                                       para=self.params.knn_distance + ':' + str(
                                                           self.params.k), adjTag=(
                            self.params.useGAEembedding or self.params.useBothembedding))

        if self.debugMode == 'savePrune':
            # Add protocol=4 for serizalize object larger than 4GiB
            with open('edgeListFile', 'wb') as edgeListFile:
                pkl.dump(self.edgeList, edgeListFile, protocol=4)

            with open('adjFile', 'wb') as adjFile:
                pkl.dump(self.adj, adjFile, protocol=4)

            with open('zOutFile', 'wb') as zOutFile:
                pkl.dump(zOut, zOutFile, protocol=4)

            with open('reconFile', 'wb') as reconFile:
                pkl.dump(recon, reconFile, protocol=4)

            with open('originalFile', 'wb') as originalFile:
                pkl.dump(original, originalFile, protocol=4)

            sys.exit(0)

        if self.params.debugMode == 'loadPrune':
            with open('edgeListFile', 'rb') as edgeListFile:
                self.edgeList = pkl.load(edgeListFile)

            with open('adjFile', 'rb') as adjFile:
                self.adj = pkl.load(adjFile)

            with open('zOutFile', 'rb') as zOutFile:
                zOut = pkl.load(zOutFile)

            with open('reconFile', 'rb') as reconFile:
                recon = pkl.load(reconFile)

            with open('originalFile', 'rb') as originalFile:
                original = pkl.load(originalFile)

        if self.params.useGAEembedding or self.params.useBothembedding:
            zDiscret = zOut > np.mean(zOut, axis=0)
            zDiscret = 1.0 * zDiscret
            if self.params.useGAEembedding:
                # mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
                # print('Mem consumption: '+str(mem))
                zOut = self.GAEembedding(zDiscret, self.adj, self.params)
                print("---GAE embedding finished")
                # mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
                # print('Mem consumption: '+str(mem))

            elif self.params.useBothembedding:
                zEmbedding = self.GAEembedding(zDiscret, self.adj, self.params)
                zOut = np.concatenate((zOut, zEmbedding), axis=1)

        G0 = nx.Graph()
        G0.add_weighted_edges_from(self.edgeList)
        nlG0 = nx.normalized_laplacian_matrix(G0)
        # set iteration criteria for converge
        adjOld = nlG0
        # set celltype criteria for converge
        listResultOld = [1 for i in range(zOut.shape[0])]

        # Fill the zeros before EM iteration
        # need better implementation later
        # this parameter largely unhelpful
        # broken code anyways
        # if self.params.zerofillFlag:
        #     for nz_index in range(len(scData.nz_i)):
        #         # tmp = scipy.sparse.lil_matrix.todense(scData.features[scData.nz_i[nz_index], scData.nz_j[nz_index]])
        #         # tmp = np.asarray(tmp).reshape(-1)[0]
        #         tmp = scData.features[scData.nz_i[nz_index], scData.nz_j[nz_index]]
        #         reconOut[scData.nz_i[nz_index], scData.nz_j[nz_index]] = tmp
        #     recon = reconOut

        # Define resolution
        # Default: auto, otherwise use user defined resolution
        if self.params.resolution == 'auto':
            if zOut.shape[0] < 2000:
                resolution = 0.8
            else:
                resolution = 0.5
        else:
            resolution = float(self.params.resolution)

        print("---EM process starts")

        for bigepoch in range(0, self.params.EM_iteration):
            print('---Start %sth iteration.' % (bigepoch))
            # Now for both methods, we need do clustering, using clustering results to check converge
            # Clustering: Get clusters
            if self.params.clustering_method == 'Louvain':
                listResult, size = self.generateLouvainCluster(self.edgeList)
                k = len(np.unique(listResult))
                print('Louvain cluster: ' + str(k))
            elif self.params.clustering_method == 'LouvainK':
                listResult, size = self.generateLouvainCluster(self.edgeList)
                k = len(np.unique(listResult))
                print('Louvain cluster: ' + str(k))
                k = int(k * resolution) if int(k * resolution) >= 3 else 2
                clustering = KMeans(n_clusters=k, random_state=0).fit(zOut)
                # clustering = MiniBatchKMeans(n_clusters=k, random_state=0, batch_size=128).fit(zOut)
                listResult = clustering.predict(zOut)
            elif self.params.clustering_method == 'LouvainB':
                listResult, size = self.generateLouvainCluster(self.edgeList)
                k = len(np.unique(listResult))
                print('Louvain cluster: ' + str(k))
                k = int(k * resolution) if int(k * resolution) >= 3 else 2
                clustering = Birch(n_clusters=k).fit(zOut)
                listResult = clustering.predict(zOut)
            elif self.params.clustering_method == 'KMeans':
                clustering = KMeans(n_clusters=self.params.n_clusters,
                                    random_state=0).fit(zOut)
                listResult = clustering.predict(zOut)
            elif self.params.clustering_method == 'SpectralClustering':
                clustering = SpectralClustering(
                    n_clusters=self.params.n_clusters, assign_labels="discretize", random_state=0).fit(zOut)
                listResult = clustering.labels_.tolist()
            elif self.params.clustering_method == 'AffinityPropagation':
                clustering = AffinityPropagation().fit(zOut)
                listResult = clustering.predict(zOut)
            elif self.params.clustering_method == 'AgglomerativeClustering':
                clustering = AgglomerativeClustering().fit(zOut)
                listResult = clustering.labels_.tolist()
            elif self.params.clustering_method == 'AgglomerativeClusteringK':
                clustering = AgglomerativeClustering(
                    n_clusters=self.params.n_clusters).fit(zOut)
                listResult = clustering.labels_.tolist()
            elif self.params.clustering_method == 'Birch':
                clustering = Birch(n_clusters=self.params.n_clusters).fit(zOut)
                listResult = clustering.predict(zOut)
            elif self.params.clustering_method == 'BirchN':
                clustering = Birch(n_clusters=None).fit(zOut)
                listResult = clustering.predict(zOut)
            elif self.params.clustering_method == 'MeanShift':
                clustering = MeanShift().fit(zOut)
                listResult = clustering.predict(zOut)
            elif self.params.clustering_method == 'OPTICS':
                clustering = OPTICS(min_samples=int(
                    self.params.k / 2), min_cluster_size=self.params.minMemberinCluster).fit(zOut)
                listResult = clustering.predict(zOut)
            else:
                print("Error: Clustering method not appropriate")
            print("---Clustering Ends")

            # If clusters more than maxclusters, then have to stop
            if len(set(listResult)) > self.params.maxClusterNumber or len(set(listResult)) <= 1:
                print("Stopping: Number of clusters is " +
                      str(len(set(listResult))) + ".")
                # Exit
                # return None
                # Else: dealing with the number
                listResult = self.trimClustering(
                    listResult, minMemberinCluster=self.params.minMemberinCluster,
                    maxClusterNumber=self.params.maxClusterNumber)

            # Debug: Calculate silhouette
            # measure_clustering_results(zOut, listResult)
            print('Total Cluster Number: ' + str(len(set(listResult))))
            # mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
            # print('Mem consumption: '+str(mem))

            # Graph regulizated EM AE with Cluster AE, do the additional AE
            if not self.params.quickmode:
                # Each cluster has a autoencoder, and organize them back in iteraization
                print('---Start Cluster Autoencoder.')
                clusterIndexList = []
                for i in range(len(set(listResult))):
                    clusterIndexList.append([])
                for i in range(len(listResult)):
                    assignee = listResult[i]
                    # Avoid bugs for maxClusterNumber
                    if assignee == self.params.maxClusterNumber:
                        assignee = self.params.maxClusterNumber - 1
                    clusterIndexList[assignee].append(i)

                reconNew = np.zeros(
                    (scData.features.shape[0], scData.features.shape[1]))

                # Convert to Tensor
                reconNew = torch.from_numpy(reconNew)
                if self.params.precisionModel == 'Double':
                    reconNew = reconNew.type(torch.DoubleTensor)
                elif self.params.precisionModel == 'Float':
                    reconNew = reconNew.type(torch.FloatTensor)
                reconNew = reconNew.to(self.device)

                model.load_state_dict(ptstatus)

                for clusterIndex in clusterIndexList:
                    reconUsage = recon[clusterIndex]
                    scDataInter = ScDatasetInter(reconUsage)
                    train_loader = DataLoader(
                        scDataInter, batch_size=self.batch_size, shuffle=False, **self.kwargs)

                    for epoch in range(1, self.cluster_epochs + 1):
                        reconCluster, originalCluster, zCluster = self.train(
                            model, optimizer, epoch, self.params, train_loader, EMFlag=True)
                    count = 0
                    for i in clusterIndex:
                        reconNew[i] = reconCluster[count, :]
                        count += 1
                    # empty cuda cache
                    del originalCluster
                    del zCluster
                    torch.cuda.empty_cache()

                # Update
                recon = reconNew
                ptstatus = model.state_dict()

                # Debug mem consumption
                # mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
                # print('Mem consumption: '+str(mem))

            # Use new dataloader
            scDataInter = ScDatasetInter(recon.detach().to('cpu'))
            train_loader = DataLoader(
                scDataInter, batch_size=self.batch_size, shuffle=False, **self.kwargs)

            for epoch in range(1, self.EM_epochs + 1):
                recon, original, z = self.train(model, optimizer, epoch, self.params, train_loader, EMFlag=True)
                # recon, original, z = train(epoch, train_loader=train_loader, EMFlag=True)

            zOut = z.detach().cpu().numpy()

            # mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
            # print('Mem consumption: '+str(mem))
            print('---Start Prune')

            self.adj, self.edgeList = scGNNgenerateAdj(zOut, graphType=self.params.prunetype,
                                                       para=self.params.knn_distance + ':' + str(
                                                           self.params.k), adjTag=(
                            self.params.useGAEembedding or self.params.useBothembedding or (
                                bigepoch == int(self.params.EM_iteration) - 1)))

            print('---Prune Finished')
            # mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
            # print('Mem consumption: '+str(mem))

            # Whether use GAE embedding
            if self.params.useGAEembedding or self.params.useBothembedding:
                zDiscret = zOut > np.mean(zOut, axis=0)
                zDiscret = 1.0 * zDiscret
                if self.params.useGAEembedding:
                    # mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
                    # print('Mem consumption: '+str(mem))
                    zOut = self.GAEembedding(zDiscret, self.adj, self.params)
                    print("---GAE embedding finished")
                    # mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
                    # print('Mem consumption: '+str(mem))
                elif self.params.useBothembedding:
                    zEmbedding = self.GAEembedding(zDiscret, self.adj, self.params)
                    zOut = np.concatenate((zOut, zEmbedding), axis=1)

            # Original save step by step
            if self.params.saveinternal:
                print('---Start save internal results')
                reconOut = recon.detach().cpu().numpy()

                # Output
                print('---Prepare save')
                # print('Save results with reconstructed shape:'+str(reconOut.shape)+' Size of gene:'+str(len(genelist))+' Size of cell:'+str(len(celllist)))
                recon_df = pd.DataFrame(np.transpose(
                    reconOut), index=self.genelist, columns=self.celllist)
                recon_df.to_csv(
                    self.params.outputDir + self.params.datasetName + '_' + self.params.regulized_type + '_' + str(
                        self.params.alphaRegularizePara) + '_' + str(self.params.L1Para) + '_' + str(
                        self.params.L2Para) + '_recon_' + str(
                        bigepoch) + '.csv')
                emblist = []
                for i in range(zOut.shape[1]):
                    emblist.append('embedding' + str(i))
                embedding_df = pd.DataFrame(zOut, index=self.celllist, columns=emblist)
                embedding_df.to_csv(
                    self.params.outputDir + self.params.datasetName + '_' + self.params.regulized_type + '_' + str(
                        self.params.alphaRegularizePara) + '_' + str(self.params.L1Para) + '_' + str(
                        self.params.L2Para) + '_embedding_' + str(
                        bigepoch) + '.csv')
                graph_df = pd.DataFrame(
                    self.edgeList, columns=["NodeA", "NodeB", "Weights"])
                graph_df.to_csv(
                    self.params.outputDir + self.params.datasetName + '_' + self.params.regulized_type + '_' + str(
                        self.params.alphaRegularizePara) +
                    '_' + str(self.params.L1Para) + '_' + str(self.params.L2Para) + '_graph_' + str(bigepoch) + '.csv',
                    index=False)
                results_df = pd.DataFrame(
                    listResult, index=self.celllist, columns=["Celltype"])
                results_df.to_csv(
                    self.params.outputDir + self.params.datasetName + '_' + self.params.regulized_type + '_' + str(
                        self.params.alphaRegularizePara) + '_' + str(self.params.L1Para) + '_' + str(
                        self.params.L2Para) + '_results_' + str(
                        bigepoch) + '.txt')

                print('---Save internal completed')

            # mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
            # print('Mem consumption: '+str(mem))
            print('---Start test converge condition')

            # Iteration usage
            # If not only use 'celltype', we have to use graph change
            # The problem is it will consume huge memory for giant graphs
            if not self.params.converge_type == 'celltype':
                Gc = nx.Graph()
                Gc.add_weighted_edges_from(self.edgeList)
                adjGc = nx.adjacency_matrix(Gc)

                # Update new adj
                adjNew = self.params.alpha * nlG0 + \
                         (1 - self.params.alpha) * adjGc / np.sum(adjGc, axis=0)

                # mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
                # print('Mem consumption: '+str(mem))
                print('---New adj ready')

                # debug
                graphChange = np.mean(abs(adjNew - adjOld))
                graphChangeThreshold = self.params.converge_graphratio * \
                                       np.mean(abs(nlG0))
                print('adjNew:{} adjOld:{} G0:{}'.format(adjNew, adjOld, nlG0))
                print('mean:{} threshold:{}'.format(
                    graphChange, graphChangeThreshold))

                # Update
                adjOld = adjNew

            # Check similarity
            ari = adjusted_rand_score(listResultOld, listResult)

            # Debug Information of clustering results between iterations
            # print(listResultOld)
            # print(listResult)
            print('celltype similarity:' + str(ari))

            # graph criteria
            if self.params.converge_type == 'graph':
                if graphChange < graphChangeThreshold:
                    print('Converge now!')
                    break
            # celltype criteria
            elif self.params.converge_type == 'celltype':
                if ari > self.params.converge_celltyperatio:
                    print('Converge now!')
                    break
            # if both criteria are meets
            elif self.params.converge_type == 'both':
                if graphChange < graphChangeThreshold and ari > self.params.converge_celltyperatio:
                    print('Converge now!')
                    break
            # if either criteria are meets
            elif self.params.converge_type == 'either':
                if graphChange < graphChangeThreshold or ari > self.params.converge_celltyperatio:
                    print('Converge now!')
                    break

            # Update
            listResultOld = listResult
            self.listResult = listResult

            print("---" + str(bigepoch) + "th iteration in EM Finished")
            self.model = model
            self.optimizer = optimizer
            self.ptfileStart = ptfileStart
            self.zOut = zOut
            return reconOri

    def predict(self, test_data):
        """ Predict function for imputation
        Parameters
        ------------
        reconOri :
            input to be imputed

        Returns
        ------------
        embedding_df :
            dataframe embedding of
        graph_df :
            dataframe of edge list cell graph
        results_df :
            dataframe of cell types
        recon_df :
            dataframe of reconstructed input, completed imputed values
        """
        print("---Starts Imputation")
        test_data = torch.Tensor(test_data).to(self.device)
        stateStart = torch.load(self.ptfileStart)
        self.model.load_state_dict(stateStart['state_dict'])
        self.optimizer.load_state_dict(stateStart['optimizer'])

        if self.model_type == 'VAE':
            recon, _, _, z = self.model(test_data)
        elif self.model_type == 'AE':
            recon, z = self.model(test_data)

        zOut = z.detach().cpu().numpy()
        reconOut = recon.detach().cpu().numpy()
        if not self.params.noPostprocessingTag:
            threshold_indices = reconOut < self.params.postThreshold
            reconOut[threshold_indices] = 0.0

        return reconOut

    def score(self, true_expr, imputed_expr, test_idx, metric, true_labels=None, n_neighbors=None, n_pcs=None,
              clu_resolution=1, targetgenes=None):

        """Evaluate the trained model.
        Parameters
        ----------
        true_expr : DataFrame
            true expression
        imputed_expr : DataFrame
            imputed expression
        metric : string
            choice of metric, either MSE or ARI
        true_labels : array optional
            provided cell labels
        n_neighbors : int optional
            number of neighbors tp cluster imputed data
        n_pcs: int optional
            number of principal components for neighbor detection
        clu_resolution : int optional
            resolution for Leiden cluster
        targetgenes: array optional
            genes to be imputed
        Returns
        -------
        acc : float
            accuracy.
        nmi : float
            normalized mutual information.
        ari : float
            adjusted Rand index.
        mse : float
            mean squared erros
        """
        if sp.issparse(true_expr):
            true_expr = true_expr.toarray()
        # if imputed_expr == None:
        #     imputed_expr = self.predict(self.reconOri)
        if (targetgenes == None):
            targetgenes = self.genelist
        # targetgenes = targetgenes.flatten()
        allowd_metrics = {"MSE", "ARI"}
        if metric not in allowd_metrics:
            raise ValueError("scoring metric %r." % allowd_metrics)
        # imputed_expr = imputed_expr.loc[:,targetgenes] # subset target genes only
        if (metric == 'MSE'):
            true_target = true_expr[test_idx,]
            imputed_target = imputed_expr[test_idx,]
            # true_expr = true_expr.loc[imputed_expr.index, imputed_expr.columns]
            mse_cells = pd.DataFrame(((true_target - imputed_target) ** 2).mean(axis=1)).dropna()
            mse_genes = pd.DataFrame(((true_target - imputed_target) ** 2).mean(axis=0)).dropna()
            return mse_cells, mse_genes
        elif (metric == 'ARI'):
            adata = ad.AnnData(imputed_expr)
            sc.pp.normalize_total(adata, target_sum=1e4)
            sc.pp.log1p(adata)
            sc.pp.highly_variable_genes(adata)
            adata = adata[:, adata.var.highly_variable]
            sc.pp.scale(adata)
            sc.tl.pca(adata, svd_solver='arpack')
            sc.pp.neighbors(adata, n_neighbors=n_neighbors, n_pcs=n_pcs)
            sc.tl.umap(adata)
            sc.tl.leiden(adata, resolution=clu_resolution)
            pred_labels = adata.obs['leiden']
            ari = round(adjusted_rand_score(true_labels, pred_labels), 3)
            return ari



import warnings

warnings.filterwarnings("ignore")
import argparse
from argparse import Namespace
import scipy as sp
import numpy as np
import pandas as pd
import scanpy as sc
from time import time as get_time

from dance.modules.single_modality.clustering.scdsc import *
from dance.transforms.graph_construct import construct_graph_sc
from dance.transforms.preprocess import normalize_adata, filter_data
from dance.datasets.singlemodality import *
from dance.utils import set_seed

# for repeatability
set_seed(42)

if __name__ == "__main__":

    time_start = get_time()
    parser = argparse.ArgumentParser()

    # model_para = [n_enc_1(n_dec_3), n_enc_2(n_dec_2), n_enc_3(n_dec_1)]
    model_para = [512, 256, 256]
    # Cluster_para = [n_z1, n_z2, n_z3, n_init, n_input, n_clusters]
    Cluster_para = [256, 128, 32, 20, 100, 10]
    # Balance_para = [binary_crossentropy_loss, ce_loss, re_loss, zinb_loss, sigma]
    Balance_para = [1, 0.01, 0.1, 0.1, 1]

    parser.add_argument('--name', type=str,
                        default='worm_neuron_cell')  # choice=['10X_PBMC', 'mouse_bladder_cell', 'mouse_ES_cell', 'worm_neuron_cell']
    parser.add_argument('--pretrain_path', type=str, default='worm_neuron_cell')
    parser.add_argument('--graph_path', type=str, default='worm_neuron_cell')
    parser.add_argument('--method', type=str, default='p')  # choice=['heat', 'cos', 'ncos', 'p']
    parser.add_argument('--batch_size', default=256, type=int)
    parser.add_argument('--n_enc_1', default=model_para[0], type=int)
    parser.add_argument('--n_enc_2', default=model_para[1], type=int)
    parser.add_argument('--n_enc_3', default=model_para[2], type=int)
    parser.add_argument('--n_dec_1', default=model_para[2], type=int)
    parser.add_argument('--n_dec_2', default=model_para[1], type=int)
    parser.add_argument('--n_dec_3', default=model_para[0], type=int)
    parser.add_argument('--topk', type=int, default=50)
    parser.add_argument('--lr', type=float, default=1e-2)
    parser.add_argument('--pretrain_epochs', type=int, default=200)
    parser.add_argument('--n_epochs', type=int, default=500)
    parser.add_argument('--n_z1', default=Cluster_para[0], type=int)
    parser.add_argument('--n_z2', default=Cluster_para[1], type=int)
    parser.add_argument('--n_z3', default=Cluster_para[2], type=int)
    parser.add_argument('--n_input', type=int, default=Cluster_para[4])
    parser.add_argument('--n_clusters', type=int, default=Cluster_para[5])
    parser.add_argument('--device', type=str, default='cpu')
    parser.add_argument('--v', type=int, default=1)
    parser.add_argument('--nb_genes', type=int, default=5000)
    parser.add_argument('--binary_crossentropy_loss', type=float, default=Balance_para[0])
    parser.add_argument('--ce_loss', type=float, default=Balance_para[1])
    parser.add_argument('--re_loss', type=float, default=Balance_para[2])
    parser.add_argument('--zinb_loss', type=float, default=Balance_para[3])
    parser.add_argument('--sigma', type=float, default=Balance_para[4])

    args = parser.parse_args()
    args.n_input = args.nb_genes
    args.graph_path = args.name + '_' + f'{args.topk}'
    # File = [gene_expresion data file, Graph file, h5 file, pretrain_path]
    File = ['./data/' + args.name, "./graph/" + args.graph_path + "_graph.txt", "./data/" + args.name + ".h5",
            "./model/" + args.name + "_pre.pkl"]
    args.graph_path = File[1]
    args.pretrain_path = File[3]

    if not os.path.exists("./graph/"):
        os.makedirs("./graph/")

    if not os.path.exists("./model/"):
        os.makedirs("./model/")

    data = ClusteringDataset('./data', args.name).load_data()
    X = data.X
    Y = data.Y
    genes_idx, cells_idx = filter_data(X, highly_genes=args.nb_genes)
    X = X[cells_idx][:, genes_idx]
    Y = Y[cells_idx]

    adata = sc.AnnData(X, dtype=np.float32)
    adata.obs['Group'] = Y
    adata = adata.copy()
    adata.obs['DCA_split'] = 'train'
    adata.obs['DCA_split'] = adata.obs['DCA_split'].astype('category')

    adata = normalize_adata(adata,
                            size_factors=True,
                            normalize_input=True,
                            logtrans_input=True)
    X = adata.X
    Y = adata.obs['Group']
    args.n_clusters = len(np.unique(Y))
    construct_graph_sc(File[1], X, Y, args.method, args.topk)

    model = SCDSCWrapper(Namespace(**vars(args)))
    # pretrain AE
    if not os.path.exists(args.pretrain_path):
        print('Pretrain:')
        dataset_pre = PretrainDataset(X)
        model.pretrain_ae(dataset_pre, args.batch_size, args.pretrain_epochs,
                          args.n_clusters, args.pretrain_path)

    # train scDSC
    print('Train:')
    dataset = TrainingDataset(X, Y)
    X_raw = adata.raw.X
    sf = adata.obs.size_factors

    model.fit(dataset, X_raw, sf, args.graph_path, lr=args.lr, n_epochs=args.n_epochs,
              bcl=args.binary_crossentropy_loss, cl=args.ce_loss, rl=args.re_loss, zl=args.zinb_loss)
    print("Running Time：%d seconds", get_time() - time_start)

    y_pred = model.predict()
    #    print(f'Prediction: {y_pred}')

    acc, nmi, ari = model.score(Y)
    print("ACC: %.4f, NMI: %.4f, ARI: %.4f" % (acc, nmi, ari))

"""Reproduction information
10X PBMC:
python scdsc.py --name='10X_PBMC' --method='cos' --topk=30 --v=7 --binary_crossentropy_loss=0.75 --ce_loss=0.5 --re_loss=0.1 --zinb_loss=2.5 --sigma=0.4

Mouse Bladder:
python scdsc.py --name='mouse_bladder_cell' --method='p' --topk=50 --v=7 --binary_crossentropy_loss=2.5 --ce_loss=0.1 --re_loss=0.5 --zinb_loss=1.5 --sigma=0.6

Mouse ES:
python scdsc.py --name='mouse_ES_cell' --method='heat' --topk=50 --v=7 --binary_crossentropy_loss=0.1 --ce_loss=0.01 --re_loss=1.5 --zinb_loss=0.5 --sigma=0.1
    
Worm Neuron:
python scdsc.py --name='worm_neuron_cell' --method='p' --topk=20 --v=7 --binary_crossentropy_loss=2 --ce_loss=2 --re_loss=3 --zinb_loss=0.1 --sigma=0.4
"""


import torch
import argparse
import random
import numpy as np
from dance.datasets.multimodality import ModalityMatchingDataset
from dance.utils import set_seed
from dance.modules.multi_modality.match_modality.scmm import MMVAE



if __name__ == '__main__':
    rndseed = random.randint(0, 2147483647)
    parser = argparse.ArgumentParser()
    #TODO: remove config?
    parser.add_argument('--config', type=str, default = './match_modality/configs/cross_modal_supervised.yaml', help='Path to the config file.')
    parser.add_argument('--output_path', type=str, default='./modality_matching/output', help="outputs path")
    parser.add_argument('-d', '--data_folder', default='./data/modality_matching')
    parser.add_argument("--resume", action="store_true")
    parser.add_argument('-t', '--subtask', default='openproblems_bmmc_cite_phase2_rna')
    parser.add_argument('-device', '--device', default='cuda:3')
    parser.add_argument('-cpu', '--cpus', default = 1, type = int)
    parser.add_argument('-seed', '--rnd_seed', default = rndseed, type=int)


    parser.add_argument('--experiment', type=str, default='test', metavar='E',
                        help='experiment name')
    parser.add_argument('--obj', type=str, default='m_elbo_naive_warmup', metavar='O',
                        help='objective to use (default: elbo)')
    parser.add_argument('--llik_scaling', type=float, default=1.,
                        help='likelihood scaling for cub images/svhn modality when running in'
                             'multimodal setting, set as 0 to use default value')
    parser.add_argument('--batch_size', type=int, default=64, metavar='N',
                        help='batch size for data (default: 256)')
    parser.add_argument('--epochs', type=int, default=100, metavar='E',
                        help='number of epochs to train (default: 100)')
    parser.add_argument('--lr', type=float, default=1e-4, metavar='L',
                        help='learning rate (default: 1e-3)')
    parser.add_argument('--latent_dim', type=int, default=10, metavar='L',
                        help='latent dimensionality (default: 20)')
    parser.add_argument('--num_hidden_layers', type=int, default=2, metavar='H',
                        help='number of hidden layers in enc and dec (default: 2)')
    parser.add_argument('--r_hidden_dim', type=int, default=100,
                        help='number of hidden units in enc/dec for gene')
    parser.add_argument('--p_hidden_dim', type=int, default=20,
                        help='number of hidden units in enc/dec for protein/peak')
    parser.add_argument('--pre_trained', type=str, default="",
                        help='path to pre-trained model (train from scratch if empty)')
    parser.add_argument('--learn_prior', type=bool, default=True,
                        help='learn model prior parameters')
    parser.add_argument('--print_freq', type=int, default=0, metavar='f',
                        help='frequency with which to print stats (default: 0)')
    parser.add_argument('--deterministic_warmup', type=int, default=50, metavar='W',
                        help='deterministic warmup')
    args = parser.parse_args()

    torch.set_num_threads(args.cpus)
    rndseed = args.rnd_seed
    set_seed(rndseed)
    subtask = args.subtask
    dataset = ModalityMatchingDataset(subtask, data_dir=args.data_folder).load_data().load_sol().preprocess(kind='feature_selection')
    device = args.device

    args.r_dim = dataset.modalities[0].shape[1]
    args.p_dim = dataset.modalities[1].shape[1]

    model_class = 'rna-protein' if subtask == 'openproblems_bmmc_cite_phase2_rna' else 'rna-dna'
    model = MMVAE(model_class, args).to(device)

    test_X = torch.from_numpy(dataset.numpy_features(2, True)).to(device).float()
    test_Y = torch.from_numpy(dataset.numpy_features(3, True)).to(device).float()
    labels = torch.from_numpy(dataset.test_sol.X.toarray())
    model.fit(dataset)
    print(model.predict([test_X, test_Y]))
    print(model.score(test_X, test_Y, labels))







import dgl.nn as dglnn
import dgl
import torch
import torch.nn.functional as F
import torch.nn as nn
import numpy as np
import math
import copy
from torch.utils.data import DataLoader
from dance.utils import SimpleIndexDataset

class ScMoGCNWrapper(object):
    """ScMoGCN class.

    Parameters
    ----------
    args : argparse.Namespace
        A Namespace object that contains arguments of ScMoGCN. For details of parameters in parser args, please refer to link (parser help document).
    """
    def __init__(self, args):
        super(ScMoGCNWrapper, self).__init__()
        self.args = args
        self.model = ScMoGCN(args).to(args.device)

    def predict(self, graph, idx=None, device='gpu'):
        """Predict function to get latent representation of data.

        Parameters
        ----------
        graph : dgl.DGLGraph
            Cell-feature graph contructed from the dataset.
        idx : Iterable[int] optional
            Cell indices for prediction, by default to be None, where all the cells to be predicted.
        device : str optional
            Well to perform predicting, by default to be 'gpu'.

        Returns
        -------
        pred : torch.Tensor
            Predicted target modality features.

        """
        if self.args.device != 'cpu' and device=='cpu':
            model = copy.deepcopy(self.model)
            model.to('cpu')
            graph = graph.to('cpu')
        else:
            model = self.model
        model.eval()
        with torch.no_grad():
            if idx is None:
                pred = model.forward(graph)
            else:
                pred = model.forward(graph)[idx]
        pred = pred.to(self.args.device)
        return pred

    def score(self, g, idx, labels, device='gpu'):
        """Score function to get score of prediction.

        Parameters
        ----------
        g : dgl.DGLGraph
            Cell-feature graph contructed from the dataset.
        idx : Iterable[int] optional
            Index of testing cells for scoring.
        labels : torch.Tensor
            Ground truth label of cells, a.k.s target modality features.
        device : str optional
            Well to perform predicting, by default to be 'gpu'.

        Returns
        -------
        loss : float
            RMSE loss of predicted output modality features.

        """
        self.model.eval()
        with torch.no_grad():
            logits = F.relu(self.predict(g, idx, device))
            loss = math.sqrt(F.mse_loss(logits, labels).item())
            return loss

    # TODO: need to modify the logic of validation and test to adapt Inductive learning;
    #  w. test = Transductive learning, w/o = Inductive learning
    def fit(self, g, y, split=None, eval=True, verbose=2, y_test=None, logger=None, sampling=False):
        """fit function for training.

        Parameters
        ----------
        g : dgl.DGLGraph
            Cell-feature graph contructed from the dataset.
        y : torch.Tensor
            Labels of each training cell, a.k.a target modality features.
        split : dictionary optional
            Cell indices for train-test split, needed when eval parameter set to be True.
        eval : bool optional
            Whether to evaluate during training, by default to be True.
        verbose : int optional
            Verbose level, by default to be 2 (i.e. print and logger).
        y_test : torch.Tensor optional
            Labels of each testing cell, needed when eval parameter set to be True.
        logger : file-object optional
            Log file, needed when verbose set to be 2.
        sampling : bool optional
            Whether perform feature and cell sampling, by default to be False.

        Returns
        -------
        None.
        """
        if sampling:
            return self.fit_with_sampling(g, y, split, eval, verbose, y_test, logger)
        kwargs = vars(self.args)
        PREFIX = kwargs['prefix']
        CELL_SIZE = kwargs['CELL_SIZE']
        TRAIN_SIZE = kwargs['TRAIN_SIZE']

        if verbose > 1 and logger is None:
            logger = open(f'{kwargs["log_folder"]}/{PREFIX}.log', 'w')
        if verbose > 1:
            logger.write(str(self.model) + '\n')
            logger.flush()

        opt = torch.optim.AdamW(self.model.parameters(), lr=kwargs['learning_rate'], weight_decay=kwargs['weight_decay'])
        criterion = nn.MSELoss()
        val = []
        tr = []
        te = []
        minval = 100
        minvep = -1
        BATCH_SIZE = kwargs['batch_size']

        for epoch in range(kwargs['epoch']):
            if verbose > 1:
                logger.write(f'epoch:  {epoch}\n')

            self.model.train()
            logits = self.model(g)
            loss = criterion(logits[split['train']], y[split['train']])
            running_loss = loss.item()

            opt.zero_grad()
            loss.backward()
            opt.step()

            torch.cuda.empty_cache()
            tr.append(math.sqrt(running_loss))

            val.append(self.score(g, split['valid'], y[split['valid']]))
            if verbose > 1:
                logger.write(f'training loss:  {tr[-1]}\n')
                logger.flush()
                logger.write(f'validation loss:  {val[-1]}\n')
                logger.flush()

            if eval:
                te.append(self.score(g, np.arange(TRAIN_SIZE, CELL_SIZE), y_test))
                if verbose > 1:
                    logger.write(f'testing loss:  {te[-1]}\n')
                    logger.flush()

            if val[-1] < minval:
                minval = val[-1]
                minvep = epoch
                if kwargs['save_best']:
                    torch.save(self.model, f'{kwargs["model_folder"]}/{PREFIX}.best.pth')

            if epoch > 1500 and kwargs['early_stopping'] > 0 and min(val[-kwargs['early_stopping']:]) > minval:
                if verbose > 1:
                    logger.write('Early stopped.\n')
                break

            if epoch > 1200:
                if epoch % 15 == 0:
                    for p in opt.param_groups:
                        p['lr'] *= kwargs['lr_decay']

            if verbose > 0:
                print('epoch', epoch)
                print('training: ', tr[-1])
                print('valid: ', val[-1])
                if eval:
                    print('testing: ', te[-1])

        if kwargs['save_final']:
            state = {'model': self.model, 'optimizer': opt.state_dict(), 'epoch': epoch - 1}
            torch.save(state, f'{kwargs["model_folder"]}/{PREFIX}.epoch{epoch}.pth')

        if verbose > 1:
            if eval:
                logger.write(
                    f'epoch {minvep} minimal val {minval} with training:  {tr[minvep]} and testing: {te[minvep]}\n')
            else:
                logger.write(
                    f'epoch {minvep} minimal val {minval} with training:  {tr[minvep]}\n')
            logger.close()

        if verbose > 0 and eval:
            print('min testing', min(te), te.index(min(te)))
            print('converged testing', minvep, te[minvep])

        return self.model

    def fit_with_sampling(self, g, y, split=None, eval=True, verbose=2, y_test=None, logger=None):
        """fit function for training with graph sampling.

        Parameters
        ----------
        g : dgl.DGLGraph
            Cell-feature graph contructed from the dataset.
        y : torch.Tensor
            Labels of each training cell, a.k.a target modality features.
        split : dictionary optional
            Cell indices for train-test split, needed when eval parameter set to be True.
        eval : bool optional
            Whether to evaluate during training, by default to be True.
        verbose : int optional
            Verbose level, by default to be 2 (i.e. print and logger).
        y_test : torch.Tensor optional
            Labels of each testing cell, needed when eval parameter set to be True.
        logger : file-object optional
            Log file, needed when verbose set to be 2.

        Returns
        -------
        None.
        """
        kwargs = vars(self.args)
        PREFIX = kwargs['prefix']
        CELL_SIZE = kwargs['CELL_SIZE']
        TRAIN_SIZE = kwargs['TRAIN_SIZE']
        BATCH_SIZE = kwargs['batch_size']

        if verbose > 1 and logger is None:
            logger = open(f'{kwargs["log_folder"]}/{PREFIX}.log', 'w')
        if verbose > 1:
            logger.write(str(self.model) + '\n')
            logger.flush()
        g.nodes['cell'].data['label'] = torch.cat([y, y_test], 0)
        g_origin = g
        # g = g.to('cpu')
        g = g.long()
        train_nid = torch.tensor(split['train'])#.to(self.args.device)
        # sampler = dgl.sampling.PinSAGESampler(g, 'cell')
        # sampler = dgl.dataloading.NeighborSampler([{#('feature', 'pathway', 'feature'):0,
        #                                             ('cell', 'cell2feature', 'feature'):100,
        #                                             ('feature', 'feature2cell', 'cell'):100},
        #                                            {#('feature', 'pathway', 'feature'): 0,
        #                                             ('cell', 'cell2feature', 'feature'): 100,
        #                                             ('feature', 'feature2cell', 'cell'): 100},
        #                                            {#('feature', 'pathway', 'feature'): 0,
        #                                             ('cell', 'cell2feature', 'feature'): 100,
        #                                             ('feature', 'feature2cell', 'cell'): 100},
        #                                            {#('feature', 'pathway', 'feature'): 0,
        #                                             ('cell', 'cell2feature', 'feature'): 100,
        #                                             ('feature', 'feature2cell', 'cell'): 100},],
        #
        #                                             # ('feature', 'pathway' ,'feature'): 5,}]
        #                                           prob = 'weight', output_device='cpu')

        # sampler = dgl.dataloading.SAINTSampler(mode='node', budget=6000, cache=True)
        # dataloader = dgl.dataloading.DataLoader(
        #     g,  # The graph must be on GPU.
        #     {'cell': train_nid},  # train_nid must be on GPU.
        #     sampler,
        #     device = torch.device('cpu'), #torch.device(self.args.device),  # The device argument must be GPU.
        #     num_workers=0,  # Number of workers must be 0.
        #     batch_size=1000,
        #     drop_last=False,
        #     shuffle=True)

        train_dataset = SimpleIndexDataset(split['train'])
        dataloader = DataLoader(
            dataset=train_dataset,
            batch_size=BATCH_SIZE,
            shuffle=True,
            num_workers=0,
            drop_last=True,
        )
        feature_weight = g.in_degrees(etype='cell2feature').float()  # /g.in_degrees(etype='cell2feature').sum()

        opt = torch.optim.AdamW(self.model.parameters(), lr=kwargs['learning_rate'], weight_decay=kwargs['weight_decay'])
        criterion = nn.MSELoss()
        val = []
        tr = []
        te = []
        minval = 100
        minvep = -1

        for epoch in range(kwargs['epoch']):
            if verbose > 1:
                logger.write(f'epoch:  {epoch}\n')

            self.model.train()
            running_loss = 0

            # for input_nodes, output_nodes, blocks in dataloader:

            self.model.train()
            for i, batch_idx in enumerate(dataloader):
                # feature_sampled = np.random.choice(g.nodes('feature'), 0.5*len(g.nodes('feature'), replace=False),
                #                  p=feature_weight)
                feature_sampled = torch.multinomial(feature_weight, int(self.args.node_sampling_rate*len(g.nodes('feature'))), replacement=False).to(self.args.device)
                # feature_sampled = g.nodes('feature')[feature_sampled]
                subgraph = dgl.node_subgraph(g, {'cell': batch_idx.to(self.args.device), 'feature': feature_sampled})# g.nodes('feature')})
                logits = self.model(subgraph)
                output_labels = subgraph.nodes['cell'].data['label']

                # blocks = [b.to(torch.device(self.args.device)) for b in blocks]
                # logits = self.model(blocks, sampled = True)

                # output_labels = blocks[-1].dstdata['label']['cell']

                loss = criterion(logits, output_labels)

                running_loss += loss.item()

                opt.zero_grad()
                loss.backward()
                opt.step()

                del subgraph
                del output_labels
                del loss
                torch.cuda.empty_cache()
            tr.append(math.sqrt(running_loss/len(dataloader)))

            val.append(self.score(g_origin, split['valid'], y[split['valid']], 'cpu'))

            if verbose > 1:
                logger.write(f'training loss:  {tr[-1]}\n')
                logger.flush()
                logger.write(f'validation loss:  {val[-1]}\n')
                logger.flush()

            if eval:
                te.append(self.score(g_origin, np.arange(TRAIN_SIZE, CELL_SIZE), y_test, 'cpu'))
                if verbose > 1:
                    logger.write(f'testing loss:  {te[-1]}\n')
                    logger.flush()

            if val[-1] < minval:
                minval = val[-1]
                minvep = epoch
                if kwargs['save_best']:
                    torch.save(self.model, f'{kwargs["model_folder"]}/{PREFIX}.best.pth')

            if epoch > 1500 and kwargs['early_stopping'] > 0 and min(val[-kwargs['early_stopping']:]) > minval:
                if verbose > 1:
                    logger.write('Early stopped.\n')
                break

            if epoch > 1200:
                if epoch % 15 == 0:
                    for p in opt.param_groups:
                        p['lr'] *= kwargs['lr_decay']

            if verbose > 0:
                print('epoch', epoch)
                print('training: ', tr[-1])
                print('valid: ', val[-1])
                if eval:
                    print('testing: ', te[-1])

            torch.cuda.empty_cache()

        if kwargs['save_final']:
            state = {'model': self.model, 'optimizer': opt.state_dict(), 'epoch': epoch - 1}
            torch.save(state, f'{kwargs["model_folder"]}/{PREFIX}.epoch{epoch}.pth')

        if verbose > 1:
            if eval:
                logger.write(
                    f'epoch {minvep} minimal val {minval} with training:  {tr[minvep]} and testing: {te[minvep]}\n')
            else:
                logger.write(
                    f'epoch {minvep} minimal val {minval} with training:  {tr[minvep]}\n')
            logger.close()

        if verbose > 0 and eval:
            print('min testing', min(te), te.index(min(te)))
            print('converged testing', minvep, te[minvep])

        return self.model



class ScMoGCN(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.opw = args.only_pathway
        self.npw = args.no_pathway
        self.nrc = args.no_readout_concatenate
        
        hid_feats = args.hidden_size
        out_feats = args.OUTPUT_SIZE
        FEATURE_SIZE = args.FEATURE_SIZE
        
        if not args.no_batch_features:
            self.extra_encoder = nn.Linear(args.BATCH_NUM, hid_feats)
            
        if args.cell_init == 'none':
            self.embed_cell = nn.Embedding(2, hid_feats)
        else:
            self.embed_cell = nn.Linear(100, hid_feats)
            
        self.embed_feat = nn.Embedding(FEATURE_SIZE, hid_feats)
        
        self.input_linears = nn.ModuleList()
        self.input_acts = nn.ModuleList()
        self.input_norm = nn.ModuleList()
        for i in range((args.embedding_layers-1)*2):
            self.input_linears.append(nn.Linear(hid_feats, hid_feats))
        if args.activation == 'gelu':
            for i in range((args.embedding_layers-1)*2):
                self.input_acts.append(nn.GELU())
        elif args.activation == 'prelu':
            for i in range((args.embedding_layers-1)*2):
                self.input_acts.append(nn.PReLU())
        elif args.activation == 'relu':
            for i in range((args.embedding_layers-1)*2):
                self.input_acts.append(nn.ReLU())
        elif args.activation == 'leaky_relu':
            for i in range((args.embedding_layers-1)*2):
                self.input_acts.append(nn.LeakyReLU()) 
        if args.normalization == 'batch':
            for i in range((args.embedding_layers-1)*2):
                self.input_norm.append(nn.BatchNorm1d(hid_feats))
        elif args.normalization == 'layer':
            for i in range((args.embedding_layers-1)*2):
                self.input_norm.append(nn.LayerNorm(hid_feats))
        elif args.normalization == 'group':
            for i in range((args.embedding_layers-1)*2):
                self.input_norm.append(nn.GroupNorm(4, hid_feats))
        
        if self.opw:
            self.edges = ['feature2cell', 'pathway']
        elif self.npw:
            self.edges = ['feature2cell', 'cell2feature']
        else:
            self.edges = ['feature2cell', 'cell2feature', 'pathway']
        
        self.conv_layers = nn.ModuleList()
        if args.residual == 'res_cat':
            self.conv_layers.append(dglnn.HeteroGraphConv(dict(zip(self.edges, [dglnn.SAGEConv(in_feats=hid_feats, out_feats=hid_feats, aggregator_type=args.agg_function, norm=None) for i in range(len(self.edges))])), aggregate='stack'))
            for i in range(args.conv_layers-1):
                self.conv_layers.append(dglnn.HeteroGraphConv(dict(zip(self.edges, [dglnn.SAGEConv(in_feats=hid_feats*2, out_feats=hid_feats, aggregator_type=args.agg_function, norm=None) for i in range(len(self.edges))])), aggregate='stack'))

        else:
            for i in range(args.conv_layers):
                self.conv_layers.append(dglnn.HeteroGraphConv(dict(zip(self.edges, [dglnn.SAGEConv(in_feats=hid_feats, out_feats=hid_feats, aggregator_type=args.agg_function, norm=None) for i in range(len(self.edges))])), aggregate='stack'))    
        
        self.conv_acts = nn.ModuleList()
        self.conv_norm = nn.ModuleList()
        if args.activation == 'gelu':
            for i in range(args.conv_layers*2):
                self.conv_acts.append(nn.GELU())
        elif args.activation == 'prelu':
            for i in range(args.conv_layers*2):
                self.conv_acts.append(nn.PReLU())
        elif args.activation == 'relu':
            for i in range(args.conv_layers*2):
                self.conv_acts.append(nn.ReLU())
        elif args.activation == 'leaky_relu':
            for i in range(args.conv_layers*2):
                self.conv_acts.append(nn.LeakyReLU())
            
        if args.normalization == 'batch':
            for i in range(args.conv_layers*len(self.edges)):
                self.conv_norm.append(nn.BatchNorm1d(hid_feats))
        elif args.normalization == 'layer':
            for i in range(args.conv_layers*len(self.edges)):
                self.conv_norm.append(nn.LayerNorm(hid_feats))
        elif args.normalization == 'group':
            for i in range(args.conv_layers*len(self.edges)):
                self.conv_norm.append(nn.GroupNorm(4, hid_feats))
        
        self.att_linears = nn.ModuleList()
        if args.pathway_aggregation == 'attention':
            for i in range(args.conv_layers):
                self.att_linears.append(nn.Linear(hid_feats, hid_feats))
        elif args.pathway_aggregation == 'one_gate':
            for i in range(args.conv_layers):
                self.att_linears.append(nn.Linear(hid_feats*3, hid_feats))
        elif args.pathway_aggregation == 'two_gate': 
            for i in range(args.conv_layers*2):
                self.att_linears.append(nn.Linear(hid_feats*2, hid_feats))
        elif args.pathway_aggregation == 'cat':
            for i in range(args.conv_layers):
                self.att_linears.append(nn.Linear(hid_feats*2, hid_feats))
                
        self.readout_linears = nn.ModuleList()
        self.readout_acts = nn.ModuleList()
        
        if args.weighted_sum:
            print("Weighted_sum enabled. Argument '--no_readout_concatenate' won't take effect.")
            for i in range(args.readout_layers-1):
                self.readout_linears.append(nn.Linear(hid_feats, hid_feats))
            self.readout_linears.append(nn.Linear(hid_feats, out_feats))
        elif self.nrc:
            for i in range(args.readout_layers-1):
                self.readout_linears.append(nn.Linear(hid_feats, hid_feats))
            self.readout_linears.append(nn.Linear(hid_feats, out_feats))
        else:
            for i in range(args.readout_layers-1):
                self.readout_linears.append(nn.Linear(hid_feats*args.conv_layers, hid_feats*args.conv_layers))
            self.readout_linears.append(nn.Linear(hid_feats*args.conv_layers, out_feats))
            
        if args.activation == 'gelu':
            for i in range(args.readout_layers-1):
                self.readout_acts.append(nn.GELU())
        elif args.activation == 'prelu':
            for i in range(args.readout_layers-1):
                self.readout_acts.append(nn.PReLU())
        elif args.activation == 'relu':
            for i in range(args.readout_layers-1):
                self.readout_acts.append(nn.ReLU())
        elif args.activation == 'leaky_relu':
            for i in range(args.readout_layers-1):
                self.readout_acts.append(nn.LeakyReLU())
                
        self.wt = nn.Parameter(torch.zeros(args.conv_layers))
        if args.pathway_aggregation == 'alpha' and args.pathway_alpha<0:
            self.aph = nn.Parameter(torch.zeros(2))

    def attention_agg(self, layer, h0, h):
        # h: h^{l-1}, dimension: (batch, hidden)
        # feats: result from two conv(cell conv and pathway conv), stacked together; dimension: (batch, 2, hidden)
        args = self.args
        if h.shape[1]==1:
            return self.conv_norm[layer*len(self.edges)+1](h.squeeze(1))
        elif args.pathway_aggregation == 'sum':
            return h[:, 0, :] + h[:, 1, :]
        else:
            h1 = h[:, 0, :]
            h2 = h[:, 1, :]
            
            if args.subpath_activation:
                h1 = F.leaky_relu(h1)
                h2 = F.leaky_relu(h2)
                
            h1 = self.conv_norm[layer*len(self.edges)+1](h1)
            h2 = self.conv_norm[layer*len(self.edges)+2](h2)
        
        if args.pathway_aggregation == 'attention':
            feats = torch.stack([h1, h2], 1)
            att = torch.transpose(F.softmax(torch.matmul(feats, self.att_linears[layer](h0).unsqueeze(-1)), 1), 1, 2)
            feats = torch.matmul(att, feats)
            return feats.squeeze(1)
        elif args.pathway_aggregation == 'one_gate':
            att = torch.sigmoid(self.att_linears[layer](torch.cat([h0, h1, h2], 1)))
            return att * h1 + (1-att) * h2
        elif args.pathway_aggregation == 'two_gate':
            att1 = torch.sigmoid(self.att_linears[layer*2](torch.cat([h0, h1], 1)))
            att2 = torch.sigmoid(self.att_linears[layer*2+1](torch.cat([h0, h2], 1)))
            return att1 * h1 + att2 * h2
        elif args.pathway_aggregation == 'alpha':
            if args.pathway_alpha<0:
                weight = torch.softmax(self.aph, -1)
                return weight[0] * h1 + weight[1] * h2
            else:
                return (1-args.pathway_alpha) * h1 + args.pathway_alpha * h2
        elif args.pathway_aggregation == 'cat':
            return self.att_linears[layer](torch.cat([h1, h2], 1))
        
    def conv(self, graph, layer, h, hist):
        args = self.args
        h0 = hist[-1]
        h = self.conv_layers[layer](graph, h, mod_kwargs=dict(zip(self.edges, [{'edge_weight':  F.dropout(graph.edges[self.edges[i]].data['weight'], p=args.edge_dropout, training=self.training)} for i in range(len(self.edges))])))
        
        if args.model_dropout>0:
            h = {'feature': F.dropout(self.conv_acts[layer*2](self.attention_agg(layer, h0['feature'], h['feature'])), p=args.model_dropout, training=self.training),
             'cell': F.dropout(self.conv_acts[layer*2+1](self.conv_norm[layer*len(self.edges)](h['cell'].squeeze(1))), p=args.model_dropout, training=self.training)}
        else:
            h = {'feature': self.conv_acts[layer*2](self.attention_agg(layer, h0['feature'], h['feature'])),
             'cell': self.conv_acts[layer*2+1](self.conv_norm[layer*len(self.edges)](h['cell'].squeeze(1)))}
            
        return h

    def calculate_initial_embedding(self, graph):
        args = self.args

        input1 = F.leaky_relu(self.embed_feat(graph.srcdata['id']['feature']))
        input2 = F.leaky_relu(self.embed_cell(graph.srcdata['id']['cell']))

        if not args.no_batch_features:
            batch_features = graph.srcdata['bf']['cell']
            input2 += F.leaky_relu(F.dropout(self.extra_encoder(batch_features), p=0.2, training=self.training))[
                      :input2.shape[0]]

        hfeat = input1
        hcell = input2
        for i in range(args.embedding_layers - 1, (args.embedding_layers - 1) * 2):
            hfeat = self.input_linears[i](hfeat)
            hfeat = self.input_acts[i](hfeat)
            if args.normalization != 'none':
                hfeat = self.input_norm[i](hfeat)
            if args.model_dropout > 0:
                hfeat = F.dropout(hfeat, p=args.model_dropout, training=self.training)

        for i in range(args.embedding_layers-1):
            hcell = self.input_linears[i](hcell)
            hcell = self.input_acts[i](hcell)
            if args.normalization != 'none':
                hcell = self.input_norm[i](hcell)
            if args.model_dropout>0:
                hcell = F.dropout(hcell, p=args.model_dropout, training = self.training)

        return hfeat, hcell

    def propagate_with_sampling(self, blocks):
        args = self.args
        hfeat, hcell = self.calculate_initial_embedding(blocks[0])

        h = {'feature': hfeat, 'cell': hcell}

        for i in range(args.conv_layers):

            if i > 0:
                hfeat0, hcell0 = self.calculate_initial_embedding(blocks[i])

                h = {'feature': torch.cat([h['feature'], hfeat0], 1),
                     'cell': torch.cat([h['cell'], hcell0], 1)}

            hist = [h]
            h = self.conv(blocks[i], i, h, hist)

        hist = [h] * (args.conv_layers+1)
        return hist  # , hist[-1]['feature']

    def propagate(self, graph):
        args = self.args
        hfeat, hcell = self.calculate_initial_embedding(graph)

        h = {'feature': hfeat, 'cell': hcell}
        hist = [h]
        
        for i in range(args.conv_layers):
            if i==0 or args.residual=='none':
                pass
            elif args.residual == 'res_add':
                if args.initial_residual:
                    h = {'feature': h['feature']+hist[0]['feature'],
                    'cell': h['cell']+hist[0]['cell']}

                else:
                    h = {'feature': h['feature']+hist[-2]['feature'],
                    'cell': h['cell']+hist[-2]['cell']}

            elif args.residual == 'res_cat':
                if args.initial_residual:
                    h = {'feature': torch.cat([h['feature'], hist[0]['feature']], 1),
                    'cell': torch.cat([h['cell'], hist[0]['cell']], 1)}
                else:
                    h = {'feature': torch.cat([h['feature'], hist[-2]['feature']], 1),
                    'cell': torch.cat([h['cell'], hist[-2]['cell']], 1)}

            h = self.conv(graph, i, h, hist)
            hist.append(h)
        
        return hist#, hist[-1]['feature']

    def forward(self, graph, sampled=False):
        args = self.args
        if sampled:
            hist = self.propagate_with_sampling(graph)
        else:
            hist = self.propagate(graph)

        if args.weighted_sum:
            h = 0
            weight = torch.softmax(self.wt, -1)
            for i in range(args.conv_layers):
                h += weight[i] * hist[i+1]['cell']            
        elif not self.nrc:
            h = torch.cat([i['cell'] for i in hist[1:]], 1)
        else:
            h = hist[-1]['cell']
            
        for i in range(args.readout_layers-1):
            h = self.readout_linears[i](h)
            h = F.dropout(self.readout_acts[i](h), p=args.model_dropout, training = self.training)
        h = self.readout_linears[-1](h)
        
        if args.output_relu == 'relu':
            return F.relu(h)
        elif args.output_relu == 'leaky_relu':
            return F.leaky_relu(h)

        return h